"""Language detection of a document."""
import re
from typing import Any, ClassVar, Iterable, Iterator, Optional, cast

from pydantic import Field as PydanticField
from typing_extensions import override

from ..schema import Field, Item, RichData, SignalInputType, field, span
from ..signal import TextSignal

LANG_CODE = 'lang_code'
TEXT_LEN_THRESHOLD = 40


class LangDetectionSignal(TextSignal):
  """Detects the language code in text.

  \

  Supports 55 languages returning their
  [ISO 639-1 codes](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes).
  """

  name: ClassVar[str] = 'lang_detection'
  display_name: ClassVar[str] = 'Language detection'

  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  split_by_paragraph: bool = PydanticField(
    default=False, description='Compute language scores for each paragraph.'
  )

  def _detect(self, text: str, langdetect: Any) -> Optional[str]:
    if len(text) < TEXT_LEN_THRESHOLD:
      return 'TOO_SHORT'
    try:
      return langdetect.detect(text)
    except langdetect.LangDetectException:
      return None

  @override
  def setup(self) -> None:
    try:
      import langdetect

      langdetect.DetectorFactory.seed = 42  # For consistent results.
    except ImportError:
      raise ImportError(
        'Could not import the "langdetect" python package. '
        'Please install it with `pip install langdetect`.'
      )

  @override
  def fields(self) -> Field:
    if self.split_by_paragraph:
      return field(fields=[field('string_span', fields={LANG_CODE: 'string'})])
    return field('string')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    import langdetect

    data = cast(Iterable[str], data)
    # Split on paragraphs.
    split_symbol = re.compile('(\r?\n){2,}')

    for text in data:
      if not self.split_by_paragraph:
        yield self._detect(text, langdetect)
        continue

      prev_end = 0
      result: list[Item] = []
      for m in split_symbol.finditer(text):
        start, end = m.span()
        text_span = text[prev_end:start]
        text_span = text_span.strip()
        if text_span:
          lang_code = self._detect(text_span, langdetect)
          if lang_code:
            result.append(span(prev_end, start, {LANG_CODE: lang_code}))
        prev_end = end

      # Process the last chunk.
      text_span = text[prev_end:]
      if text_span.strip():
        lang_code = self._detect(text_span, langdetect)
        if lang_code:
          result.append(span(prev_end, len(text), {LANG_CODE: lang_code}))

      yield result



"""Compute text statistics for a document."""
from typing import ClassVar, Iterable, Iterator, Optional

from typing_extensions import override

from ..schema import Field, Item, RichData, SignalInputType, field
from ..signal import TextSignal

SECRETS_KEY = 'secrets'
# Selected categories. For all categories, see:
# https://microsoft.github.io/presidio/supported_entities/

# Map from Presidio category to UI-visible category name.
PII_CATEGORIES = {
  'CREDIT_CARD': 'credit_card',
  'CRYPTO': 'crypto_address',
  'EMAIL_ADDRESS': 'email_address',
  'IBAN_CODE': 'bank_number',
  'IP_ADDRESS': 'ip_address',
  'PHONE_NUMBER': 'phone_number',
}


class PIISignal(TextSignal):
  """Find personally identifiable information (emails, phone numbers, secret keys, etc)."""

  name: ClassVar[str] = 'pii'
  display_name: ClassVar[str] = 'Personal Information (PII)'

  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  @override
  def fields(self) -> Field:
    return field(
      fields={
        **{cat: ['string_span'] for cat in PII_CATEGORIES.values()},
        SECRETS_KEY: ['string_span'],
      }
    )

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    try:
      from .pii_presidio import find_pii
      from .pii_secrets import find_secrets
    except ImportError:
      raise ImportError(
        'Could not import dependencies for the "PII" signal. '
        'Please install optional dependencies via `pip install lilac[pii]`.'
      )
    for text in data:
      if not isinstance(text, str):
        yield None
        continue

      secrets = list(find_secrets(text))
      pii_dict = find_pii(text)
      yield {**pii_dict, SECRETS_KEY: secrets}



"""Test the Markdown Extractor signal."""

from ..schema import field
from ..splitters.text_splitter_test_utils import text_to_expected_spans
from .markdown_code_block import MarkdownCodeBlockSignal


def test_markdown_code_block_fields() -> None:
  signal = MarkdownCodeBlockSignal()
  signal.setup()
  assert signal.fields() == field(fields=[field('string_span', fields={'language': 'string'})])


def test_markdown_code_block() -> None:
  signal = MarkdownCodeBlockSignal()
  signal.setup()

  text = """
I am trying to add an extra field to my model form in Django.

```python
class MyForm(forms.ModelForm):
    extra_field = forms.CharField()
    class Meta:
        model = MyModel
        widgets = {
            'extra_field': forms.Textarea(attrs={'placeholder': u'Bla bla'}),
        }
```
However, it appears that the widget definition for 'extra_field' in the Meta class is ignored.

```py
class MyForm(forms.ModelForm):
    extra_field = forms.CharField(widget=forms.Textarea())
    class Meta:
        model = MyModel
```

Could you explain why my first approach does not work and what I am doing wrong?

Here is the console output:
```
fake output
```

"""
  markdown_blocks = list(signal.compute([text]))

  expected_spans = text_to_expected_spans(
    text,
    [
      (
        """```python
class MyForm(forms.ModelForm):
    extra_field = forms.CharField()
    class Meta:
        model = MyModel
        widgets = {
            'extra_field': forms.Textarea(attrs={'placeholder': u'Bla bla'}),
        }
```""",
        {'language': 'python'},
      ),
      (
        """```py
class MyForm(forms.ModelForm):
    extra_field = forms.CharField(widget=forms.Textarea())
    class Meta:
        model = MyModel
```""",
        {'language': 'py'},
      ),
      (
        """```
fake output
```""",
        {'language': ''},
      ),
    ],
  )

  assert markdown_blocks == [expected_spans]



"""Utilities for working with datasets."""

import gc
import itertools
import json
import math
import os
import pprint
import secrets
from collections.abc import Iterable
from functools import partial
from typing import Any, Callable, Generator, Iterator, Optional, TypeVar, Union, cast

import numpy as np
import pyarrow as pa

from ..batch_utils import flatten_iter
from ..embeddings.vector_store import VectorDBIndex
from ..env import env
from ..parquet_writer import ParquetWriter
from ..schema import (
  EMBEDDING_KEY,
  PATH_WILDCARD,
  ROWID,
  SPAN_KEY,
  STRING,
  TEXT_SPAN_END_FEATURE,
  TEXT_SPAN_START_FEATURE,
  Field,
  Item,
  MapFn,
  PathKey,
  PathTuple,
  Schema,
  VectorKey,
  field,
  schema,
  schema_to_arrow_schema,
)
from ..signal import Signal
from ..utils import chunks, is_primitive, log, open_file

# The embedding write chunk sizes keeps the memory pressure lower as we iteratively write to the
# vector store. Embeddings are float32, taking up 4 bytes, so this results in ~130K * dims of RAM
# pressure.
EMBEDDINGS_WRITE_CHUNK_SIZE = 32_768


def _replace_embeddings_with_none(input: Union[Item, Item]) -> Union[Item, Item]:
  if isinstance(input, np.ndarray):
    return None
  if isinstance(input, dict):
    return {k: _replace_embeddings_with_none(v) for k, v in input.items()}
  if isinstance(input, list):
    return [_replace_embeddings_with_none(v) for v in input]

  return input


def replace_embeddings_with_none(input: Union[Item, Item]) -> Item:
  """Replaces all embeddings with None."""
  return cast(Item, _replace_embeddings_with_none(input))


def count_leafs(input: Union[Iterable, Iterator]) -> int:
  """Iterate through each element of the input, flattening each one, computing a count.

  Sum the final set of counts. This is the important iterable not to exhaust.
  """
  return len(list(flatten_iter(input)))


def _wrap_value_in_dict(input: Union[object, dict], props: PathTuple) -> Union[object, dict]:
  # If the signal produced no value, or nan, we should return None so the parquet value is sparse.
  if isinstance(input, float) and math.isnan(input):
    input = None
  for prop in reversed(props):
    input = {prop: input}
  return input


def _wrap_in_dicts(
  input: Union[object, Iterable[object]], spec: list[PathTuple]
) -> Union[object, Iterable[object]]:
  """Wraps an object or iterable in a dict according to the spec."""
  props = spec[0] if spec else tuple()
  if len(spec) == 1:
    return _wrap_value_in_dict(input, props)
  if input is None or isinstance(input, float) and math.isnan(input):
    # Return empty dict for missing inputs.
    return {}
  if isinstance(input, dict) or is_primitive(input):
    raise ValueError(
      f'The input should be a list, not a primitive. '
      f'Got input type: {type(input)}, with a wrapping spec: {spec}'
    )
  res = [_wrap_in_dicts(elem, spec[1:]) for elem in cast(Iterable, input)]
  return _wrap_value_in_dict(res, props)


def wrap_in_dicts(input: Iterable[object], spec: list[PathTuple]) -> Generator:
  """Wraps an object or iterable in a dict according to the spec."""
  return (_wrap_in_dicts(elem, spec) for elem in input)


def schema_contains_path(schema: Schema, path: PathTuple) -> bool:
  """Check if a schema contains a path."""
  current_field = cast(Field, schema)
  for path_part in path:
    if path_part == PATH_WILDCARD:
      if current_field.repeated_field is None:
        return False
      current_field = current_field.repeated_field
    else:
      if current_field.fields is None or path_part not in current_field.fields:
        return False
      current_field = current_field.fields[str(path_part)]
  return True


def create_json_map_output_schema(map_schema: Field, output_path: PathTuple) -> Schema:
  """Create a json schema describing the final output of a map function."""
  json_schema = map_schema.model_copy(deep=True)
  for path_part in reversed(output_path):
    if path_part == PATH_WILDCARD:
      json_schema = Field(repeated_field=json_schema)
    else:
      json_schema = Field(fields={path_part: json_schema})

  assert json_schema.fields, 'json_schema always has fields'
  return Schema(fields=json_schema.fields)


def create_signal_schema(
  signal: Signal, source_path: PathTuple, current_schema: Schema
) -> Optional[Schema]:
  """Create a signal schema describing the enriched fields.

  Returns None if the signal has no predefined schema.
  """
  leafs = current_schema.leafs
  # Validate that the enrich fields are actually a valid leaf path.
  if source_path not in leafs:
    raise ValueError(f'"{source_path}" is not a valid leaf path. Leaf paths: {leafs.keys()}')

  signal_schema = signal.fields()
  if not signal_schema:
    return None

  signal_schema.signal = signal.model_dump(exclude_none=True)

  enriched_schema = field(fields={signal.key(is_computed_signal=True): signal_schema})

  for path_part in reversed(source_path):
    if path_part == PATH_WILDCARD:
      enriched_schema = Field(repeated_field=enriched_schema)
    else:
      enriched_schema = Field(fields={path_part: enriched_schema})

  if not enriched_schema.fields:
    raise ValueError('This should not happen since enriched_schema always has fields (see above)')

  return schema(enriched_schema.fields.copy())


def _flat_embeddings(
  input: Union[Item, Iterable[Item]], path: PathKey = ()
) -> Iterator[tuple[PathKey, list[Item]]]:
  if (
    isinstance(input, list)
    and len(input) > 0
    and isinstance(input[0], dict)
    and EMBEDDING_KEY in input[0]
  ):
    yield path, input
  elif isinstance(input, dict):
    for k, v in input.items():
      yield from _flat_embeddings(v, path)
  elif isinstance(input, list):
    for i, v in enumerate(input):
      yield from _flat_embeddings(v, (*path, i))
  else:
    # Ignore other primitives.
    pass


def write_embeddings_to_disk(
  vector_store: str, signal_items: Iterable[Item], output_dir: str
) -> None:
  """Write a set of embeddings to disk."""
  path_embedding_items = (
    _flat_embeddings(signal_item, path=(signal_item[ROWID],)) for signal_item in signal_items
  )

  def _get_span_vectors() -> Iterator[Item]:
    for path_item in path_embedding_items:
      for path_key, embedding_items in path_item:
        if not path_key or not embedding_items:
          # Sparse embeddings may not have an embedding for every key.
          continue

        embedding_vectors: list[np.ndarray] = []
        spans: list[tuple[int, int]] = []

        for e in embedding_items:
          text_span = e[SPAN_KEY]
          vector = e[EMBEDDING_KEY]
          # We squeeze here because embedding functions can return outer dimensions of 1.
          embedding_vectors.append(vector.reshape(-1))
          spans.append((text_span[TEXT_SPAN_START_FEATURE], text_span[TEXT_SPAN_END_FEATURE]))

        yield (path_key, spans, embedding_vectors)

  span_vectors = _get_span_vectors()

  vector_index = VectorDBIndex(vector_store)
  for span_vectors_chunk in chunks(span_vectors, EMBEDDINGS_WRITE_CHUNK_SIZE):
    chunk_spans: list[tuple[PathKey, list[tuple[int, int]]]] = []
    chunk_embedding_vectors: list[np.ndarray] = []
    for path_key, spans, vectors in span_vectors_chunk:
      chunk_spans.append((path_key, spans))
      chunk_embedding_vectors.extend(vectors)

    embedding_matrix = np.array(chunk_embedding_vectors, dtype=np.float32)

    vector_index.add(chunk_spans, embedding_matrix)
    vector_index.save(output_dir)

    del embedding_matrix, chunk_embedding_vectors, chunk_spans
    gc.collect()

  del vector_index
  gc.collect()


def write_items_to_parquet(
  items: Iterable[Item],
  output_dir: str,
  schema: Schema,
  filename_prefix: str,
  shard_index: int,
  num_shards: int,
) -> str:
  """Write a set of items to a parquet file, in columnar format."""
  schema = schema.model_copy(deep=True)
  # Add a rowid column.
  schema.fields[ROWID] = Field(dtype=STRING)

  arrow_schema = schema_to_arrow_schema(schema)
  out_filename = get_parquet_filename(filename_prefix, shard_index, num_shards)
  filepath = os.path.join(output_dir, out_filename)
  f = open_file(filepath, mode='wb')
  writer = ParquetWriter(schema)
  writer.open(f)
  debug = env('DEBUG', False)
  num_items = 0
  for item in items:
    # Add a rowid column.
    if ROWID not in item:
      item[ROWID] = secrets.token_urlsafe(nbytes=12)  # 16 base64 characters.
    if debug:
      try:
        _validate(item, arrow_schema)
      except Exception as e:
        raise ValueError(f'Error validating item: {json.dumps(item)}') from e
    writer.write(item)
    num_items += 1
  writer.close()
  f.close()
  return out_filename


def _validate(item: Item, schema: pa.Schema) -> None:
  # Try to parse the item using the inferred schema.
  try:
    pa.RecordBatch.from_pylist([item], schema=schema)
  except pa.ArrowTypeError:
    log('Failed to parse arrow item using the arrow schema.')
    log('Item:')
    log(pprint.pformat(item, indent=2))
    log('Arrow schema:')
    log(schema)
    raise  # Re-raise the same exception, same stacktrace.


def get_parquet_filename(prefix: str, shard_index: int, num_shards: int) -> str:
  """Return the filename for a parquet file."""
  return f'{prefix}-{shard_index:05d}-of-{num_shards:05d}.parquet'


def _flatten_keys(
  rowid: str,
  nested_input: Iterable[Any],
  location: list[int],
  is_primitive_predicate: Callable[[object], bool],
) -> Iterator[VectorKey]:
  if (
    is_primitive_predicate(nested_input)
    or is_primitive(nested_input)
    or isinstance(nested_input, dict)
  ):
    yield (rowid, *location)
    return

  for i, input in enumerate(nested_input):
    yield from _flatten_keys(rowid, input, [*location, i], is_primitive_predicate)


def flatten_keys(
  rowids: Iterable[str],
  nested_input: Iterable[Any],
  is_primitive_predicate: Callable[[object], bool] = is_primitive,
) -> Iterator[Optional[PathKey]]:
  """Flatten the rowids of a nested input."""
  for rowid, input in zip(rowids, nested_input):
    if input is None:
      yield None
      continue
    yield from _flatten_keys(rowid, input, [], is_primitive_predicate)


Tin = TypeVar('Tin')
Tout = TypeVar('Tout')


def sparse_to_dense_compute(
  sparse_input: Iterator[Optional[Tin]], func: Callable[[Iterator[Tin]], Iterator[Tout]]
) -> Iterator[Optional[Tout]]:
  """Densifies the input before calling the provided `func` and sparsifies the output."""
  sparse_input = iter(sparse_input)
  sparse_input, sparse_input_2 = itertools.tee(sparse_input, 2)
  dense_input = cast(Iterator[Tin], filter(lambda x: x is not None, sparse_input_2))
  dense_output = iter(func(dense_input))
  for input in sparse_input:
    if input is None:
      yield None
    else:
      yield next(dense_output)


def shard_id_to_range(
  shard_id: Optional[int], shard_count: Optional[int], num_items: int
) -> tuple[int, int]:
  """Compute the start and end offsets for a shard."""
  shard_size = math.ceil(num_items / shard_count) if shard_count is not None else num_items
  shard_start_idx = min(shard_id * shard_size, num_items) if shard_id is not None else 0
  shard_end_idx = min((shard_id + 1) * shard_size, num_items) if shard_id is not None else num_items
  return (shard_start_idx, shard_end_idx)


def _cardinality_prefix(path: PathTuple) -> PathTuple:
  """Returns the cardinality prefix for a path."""
  # Find the last wildcard in the path.
  wildcard_idx = 0
  for i, path_part in enumerate(path):
    if path_part == PATH_WILDCARD:
      wildcard_idx = i
  return path[:wildcard_idx]


def paths_have_same_cardinality(path1: PathTuple, path2: PathTuple) -> bool:
  """Returns true if the paths have the same cardinality."""
  return _cardinality_prefix(path1) == _cardinality_prefix(path2)


def get_sibling_output_path(path: PathTuple, suffix: str) -> PathTuple:
  """Get the output path for a sibling column."""
  # Find the last non-wildcard in the path.
  index = 0
  for i, path_part in enumerate(path):
    if path_part != PATH_WILDCARD:
      index = i
  return (*path[:index], f'{path[index]}__{suffix}', *path[index + 1 :])


def get_common_ancestor(path1: PathTuple, path2: PathTuple) -> tuple[PathTuple, str, str]:
  """Get the path of the common ancestor of the two paths, along with the columns of the paths."""
  index = 0
  column1: str = ''
  column2: str = ''
  for i, (path_part1, path_part2) in enumerate(zip(path1, path2)):
    if path_part1 != path_part2:
      index = i
      column1 = path_part1
      column2 = path_part2
      break
  ancestor_path = path1[:index]
  return (ancestor_path, column1, column2)


def get_callable_name(map_fn: MapFn) -> str:
  """Get the name of a callable function."""
  if hasattr(map_fn, '__name__'):
    return map_fn.__name__
  if isinstance(map_fn, partial):
    return get_callable_name(map_fn.func)
  if hasattr(map_fn, 'name'):
    return getattr(map_fn, 'name')
  return 'unknown'



"""ShareGPT format."""
from typing import ClassVar

from ..dataset_format import DatasetFormat, DatasetFormatInputSelector
from ..schema import PATH_WILDCARD, Item, PathTuple, Schema, schema


def _sharegpt_selector(item: Item, conv_from: str) -> str:
  """Selector for ShareGPT."""
  # TODO(nsthorat): Make this return an array, and not pre-join with newlines.
  return '\n'.join(conv['value'] for conv in item['conversations'] if conv['from'] == conv_from)


_SYSTEM_SELECTOR = DatasetFormatInputSelector(
  name='system',
  selector=lambda item: _sharegpt_selector(item, 'system'),
)
_HUMAN_SELECTOR = DatasetFormatInputSelector(
  name='human',
  selector=lambda item: _sharegpt_selector(item, 'human'),
)
_GPT_SELECTOR = DatasetFormatInputSelector(
  name='gpt',
  selector=lambda item: _sharegpt_selector(item, 'gpt'),
)


class ShareGPT(DatasetFormat):
  """ShareGPT format."""

  name: ClassVar[str] = 'sharegpt'
  data_schema: Schema = schema(
    {
      'conversations': [
        {
          'from': 'string',
          'value': 'string',
        }
      ]
    }
  )
  title_slots: list[tuple[PathTuple, PathTuple]] = [
    (('conversations', PATH_WILDCARD, 'value'), ('conversations', PATH_WILDCARD, 'from'))
  ]

  system: ClassVar[DatasetFormatInputSelector] = _SYSTEM_SELECTOR
  human: ClassVar[DatasetFormatInputSelector] = _HUMAN_SELECTOR
  gpt: ClassVar[DatasetFormatInputSelector] = _GPT_SELECTOR

  input_selectors: ClassVar[dict[str, DatasetFormatInputSelector]] = {
    selector.name: selector for selector in [_SYSTEM_SELECTOR, _HUMAN_SELECTOR, _GPT_SELECTOR]
  }



"""Utils for routers."""

import traceback
from typing import Callable, Iterable, Optional

from fastapi import HTTPException, Request, Response
from fastapi.routing import APIRoute

from .auth import UserInfo
from .concepts.db_concept import DISK_CONCEPT_DB, DISK_CONCEPT_MODEL_DB
from .schema import Item, RichData
from .signals.concept_scorer import ConceptSignal


class RouteErrorHandler(APIRoute):
  """Custom APIRoute that handles application errors and exceptions."""

  def get_route_handler(self) -> Callable:
    """Get the route handler."""
    original_route_handler = super().get_route_handler()

    async def custom_route_handler(request: Request) -> Response:
      try:
        return await original_route_handler(request)
      except Exception as ex:
        if isinstance(ex, HTTPException):
          raise ex

        print('Route error:', request.url)
        print(traceback.format_exc())

        # wrap error into pretty 500 exception
        detail = ''.join(traceback.format_exception_only(type(ex), ex))
        raise HTTPException(status_code=500, detail=detail) from ex

    return custom_route_handler


def server_compute_concept(
  signal: ConceptSignal, examples: Iterable[RichData], user: Optional[UserInfo]
) -> list[Optional[Item]]:
  """Compute a concept from the REST endpoints."""
  # TODO(nsthorat): Move this to the setup() method in the concept_scorer.
  concept = DISK_CONCEPT_DB.get(signal.namespace, signal.concept_name, user)
  if not concept:
    raise HTTPException(
      status_code=404, detail=f'Concept "{signal.namespace}/{signal.concept_name}" was not found'
    )
  DISK_CONCEPT_MODEL_DB.sync(
    signal.namespace, signal.concept_name, signal.embedding, user=user, create=True
  )
  texts = [example or '' for example in examples]
  return list(signal.compute(texts))



"""Serves the Lilac server."""

import asyncio
import logging
import os
import webbrowser
from contextlib import asynccontextmanager
from importlib import metadata
from threading import Thread
from typing import Annotated, Any, AsyncGenerator, Optional

import uvicorn
from fastapi import APIRouter, BackgroundTasks, Depends, FastAPI, Request, Response
from fastapi.responses import FileResponse, HTMLResponse, JSONResponse, ORJSONResponse
from fastapi.routing import APIRoute
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from pydantic import BaseModel
from starlette.middleware.sessions import SessionMiddleware
from uvicorn.config import Config

from . import (
  router_concept,
  router_data_loader,
  router_dataset,
  router_dataset_signals,
  router_google_login,
  router_rag,
  router_signal,
  router_tasks,
)
from .auth import (
  AuthenticationInfo,
  ConceptAuthorizationException,
  UserInfo,
  get_session_user,
  get_user_access,
)
from .env import env, get_project_dir
from .load import load
from .project import create_project_and_set_env
from .router_utils import RouteErrorHandler
from .source import registered_sources

DIST_PATH = os.path.join(os.path.dirname(__file__), 'web')

tags_metadata: list[dict[str, Any]] = [
  {'name': 'datasets', 'description': 'API for querying a dataset.'},
  {'name': 'concepts', 'description': 'API for managing concepts.'},
  {'name': 'data_loaders', 'description': 'API for loading data.'},
  {'name': 'signals', 'description': 'API for managing signals.'},
]


def custom_generate_unique_id(route: APIRoute) -> str:
  """Generate the name for the API endpoint."""
  return route.name


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
  """Context manager for the lifespan of the application."""
  if env('LILAC_LOAD_ON_START_SERVER', False):

    def run() -> None:
      load(project_dir=get_project_dir(), overwrite=False)

    thread = Thread(target=run)
    thread.start()

  yield


app = FastAPI(
  default_response_class=ORJSONResponse,
  generate_unique_id_function=custom_generate_unique_id,
  openapi_tags=tags_metadata,
  lifespan=lifespan,
)


@app.exception_handler(ConceptAuthorizationException)
def concept_authorization_exception(
  request: Request, exc: ConceptAuthorizationException
) -> JSONResponse:
  """Return a 401 JSON response when an authorization exception is thrown."""
  message = 'Oops! You are not authorized to do this.'
  return JSONResponse(status_code=401, content={'detail': message, 'message': message})


@app.exception_handler(ModuleNotFoundError)
def module_not_found_error(request: Request, exc: ModuleNotFoundError) -> JSONResponse:
  """Return a 500 JSON response when a module fails to import because of optional imports."""
  message = 'Oops! You are missing a python dependency. ' + str(exc)
  return JSONResponse(status_code=500, content={'detail': message, 'message': message})


app.add_middleware(SessionMiddleware, secret_key=env('LILAC_OAUTH_SECRET_KEY'))

app.include_router(router_google_login.router, prefix='/google', tags=['google_login'])

v1_router = APIRouter(route_class=RouteErrorHandler)
v1_router.include_router(router_dataset.router, prefix='/datasets', tags=['datasets'])
v1_router.include_router(router_dataset_signals.router, prefix='/datasets', tags=['datasets'])
v1_router.include_router(router_concept.router, prefix='/concepts', tags=['concepts'])
v1_router.include_router(router_data_loader.router, prefix='/data_loaders', tags=['data_loaders'])
v1_router.include_router(router_signal.router, prefix='/signals', tags=['signals'])
v1_router.include_router(router_tasks.router, prefix='/tasks', tags=['tasks'])
v1_router.include_router(router_rag.router, prefix='/rag', tags=['rag'])

for source_name, source in registered_sources().items():
  if source.router:
    v1_router.include_router(source.router, prefix=f'/{source_name}', tags=[source_name])


@app.get('/auth_info')
def auth_info(user: Annotated[Optional[UserInfo], Depends(get_session_user)]) -> AuthenticationInfo:
  """Returns the user's ACL.

  NOTE: Validation happens server-side as well. This is just used for UI treatment.
  """
  auth_enabled = bool(env('LILAC_AUTH_ENABLED', False))
  return AuthenticationInfo(
    user=user,
    access=get_user_access(user),
    auth_enabled=auth_enabled,
    # See: https://huggingface.co/docs/hub/spaces-overview#helper-environment-variables
    huggingface_space_id=env('SPACE_ID', None),
  )


class ServerStatus(BaseModel):
  """Server status information."""

  version: str
  google_analytics_enabled: bool


@app.get('/status')
def status() -> ServerStatus:
  """Returns server status information."""
  return ServerStatus(
    version=metadata.version('lilac'),
    google_analytics_enabled=env('GOOGLE_ANALYTICS_ENABLED', False),
  )


@app.post('/load_config')
def load_config(background_tasks: BackgroundTasks) -> dict:
  """Loads from the lilac.yml."""
  background_tasks.add_task(load, project_dir=get_project_dir(), overwrite=False)
  return {}


app.include_router(v1_router, prefix='/api/v1')

current_dir = os.path.dirname(os.path.abspath(__file__))
templates = Jinja2Templates(directory=os.path.join(current_dir, 'templates'))


@app.get('/_data{path:path}', response_class=HTMLResponse, include_in_schema=False)
def list_files(request: Request, path: str) -> Response:
  """List files in the data directory."""
  if env('LILAC_AUTH_ENABLED', False):
    return Response(status_code=401)
  path = os.path.join(get_project_dir(), f'.{path}')
  if not os.path.exists(path):
    return Response(status_code=404)
  if os.path.isfile(path):
    return FileResponse(path)

  files = os.listdir(path)
  files_paths = sorted([(os.path.join(request.url.path, f), f) for f in files])
  return templates.TemplateResponse('list_files.html', {'request': request, 'files': files_paths})


@app.api_route('/{path_name}', include_in_schema=False)
def catch_all() -> FileResponse:
  """Catch any other requests and serve index for HTML5 history."""
  return FileResponse(path=os.path.join(DIST_PATH, 'index.html'))


# Serve static files in production mode.
app.mount('/', StaticFiles(directory=DIST_PATH, html=True, check_dir=False))


class GetTasksFilter(logging.Filter):
  """Task filter for /tasks."""

  def filter(self, record: logging.LogRecord) -> bool:
    """Filters out /api/v1/tasks/ from the logs."""
    return record.getMessage().find('/api/v1/tasks/') == -1


logging.getLogger('uvicorn.access').addFilter(GetTasksFilter())


class Server(uvicorn.Server):
  """Server that runs in a separate thread."""

  def __init__(self, config: Config) -> None:
    super().__init__(config)

    def run() -> None:
      try:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.serve())
      except RuntimeError:
        self.run()

    self.thread = Thread(target=run)

  def start(self, block: bool = True) -> None:
    """Start the server in a separate thread."""
    self.thread.start()
    if block:
      self.thread.join()

  def stop(self) -> None:
    """Stop the server."""
    self.should_exit = True
    self.thread.join()


SERVER: Optional[Server] = None


def start_server(
  host: str = '127.0.0.1',
  port: int = 5432,
  open: bool = False,
  project_dir: str = '',
  load: bool = False,
) -> Server:
  """Starts the Lilac web server.

  Args:
    host: The host to run the server on.
    port: The port to run the server on.
    open: Whether to open a browser tab upon startup.
    project_dir: The path to the Lilac project directory. If not specified, the `LILAC_PROJECT_DIR`
      environment variable will be used (this can be set from `set_project_dir`). If
      `LILAC_PROJECT_DIR` is not defined, will start in the current directory.
    load: Whether to load from the lilac.yml when the server boots up. This will diff the config
      with the fields that are computed and compute them when the server boots up.
  """
  create_project_and_set_env(project_dir)

  global SERVER
  if SERVER:
    raise ValueError('Server is already running')

  if load:
    os.environ['LILAC_LOAD_ON_START_SERVER'] = 'true'

  if open:

    @app.on_event('startup')
    def open_browser() -> None:
      webbrowser.open(f'http://{host}:{port}')

  block = False
  try:
    asyncio.get_event_loop()
  except RuntimeError:
    block = True

  config = uvicorn.Config(app, host=host, port=port, access_log=False)
  SERVER = Server(config)
  SERVER.start(block=block)

  return SERVER


def stop_server() -> None:
  """Stops the Lilac web server."""
  global SERVER
  if SERVER is None:
    return
  SERVER.stop()
  SERVER = None



"""Dictionary source."""
import json
from typing import Any, ClassVar, Iterable, Optional, cast

import duckdb
import fsspec
import pyarrow as pa
from typing_extensions import override

from ..schema import Item, arrow_schema_to_schema
from ..source import Source, SourceSchema

_TMP_FILENAME = 'tmp.jsonl'


class DictSource(Source):
  """Loads data from an iterable of dict objects."""

  name: ClassVar[str] = 'dict'

  _items: Iterable[Item]
  _source_schema: Optional[SourceSchema] = None
  _reader: Optional[pa.RecordBatchReader] = None
  _con: Optional[duckdb.DuckDBPyConnection] = None
  _fs: fsspec.AbstractFileSystem

  def __init__(self, items: Iterable[Item], **kwargs: Any):
    super().__init__(**kwargs)
    self._items = items

  @override
  def setup(self) -> None:
    assert self._items is not None, 'items must be set.'

    # Write the items to a temporary jsonl file in memory so we can read it via duckdb.
    self._fs = fsspec.filesystem('memory')
    with self._fs.open(_TMP_FILENAME, 'w') as file:
      for item in self._items:
        json.dump(item, file)
        file.write('\n')
    del self._items

    # Register the memory filesystem to query it.
    self._con = duckdb.connect(database=':memory:')
    self._con.register_filesystem(self._fs)
    self._con.execute(
      f"""
      CREATE VIEW t as (
        SELECT * FROM read_json_auto(
          'memory://{_TMP_FILENAME}',
          IGNORE_ERRORS=true,
          FORMAT='newline_delimited'
        )
      );
    """
    )

    res = self._con.execute('SELECT COUNT(*) FROM t').fetchone()
    num_items = cast(tuple[int], res)[0]

    self._reader = self._con.execute('SELECT * from t').fetch_record_batch(rows_per_batch=10_000)
    # Create the source schema in prepare to share it between process and source_schema.
    schema = arrow_schema_to_schema(self._reader.schema)
    self._source_schema = SourceSchema(fields=schema.fields, num_items=num_items)

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    assert self._source_schema is not None
    return self._source_schema

  @override
  def yield_items(self) -> Iterable[Item]:
    """Process the source request."""
    if not self._reader or not self._con:
      raise RuntimeError('JSON source is not initialized.')

    for batch in self._reader:
      yield from batch.to_pylist()

    self._reader.close()
    self._con.close()
    self._fs.rm(_TMP_FILENAME)
    del self._fs



"""Filters metadata based on an operatio. Internal use only for highlighting search results."""
from typing import Any, ClassVar, Iterable, Iterator, Optional, Union

from typing_extensions import override

from ..data.dataset import FeatureListValue, FeatureValue, FilterOp
from ..schema import Field, Item, SignalInputType, field
from ..signal import Signal


class FilterMaskSignal(Signal):
  """Find a substring in a document."""

  name: ClassVar[str] = 'metadata_search'
  display_name: ClassVar[str] = 'Metadata Search'
  input_type: ClassVar[SignalInputType] = SignalInputType.ANY

  op: FilterOp
  value: Optional[Union[FeatureValue, FeatureListValue]] = None

  @override
  def fields(self) -> Field:
    return field(dtype='boolean')

  @override
  def compute(self, values: Iterable[Any]) -> Iterator[Optional[Item]]:
    for value in values:
      if self.op != 'equals':
        raise ValueError(f'Metadata search does not support "{self.op}" op yet.')
      yield True if value == self.value else None



"""Implementation-agnostic tests of the Dataset DB API."""

from typing import ClassVar, Iterable, Iterator, Optional, cast

import numpy as np
import pytest
from typing_extensions import override

from ..config import (
  DatasetConfig,
  DatasetSettings,
  DatasetUISettings,
  EmbeddingConfig,
  SignalConfig,
)
from ..schema import (
  EMBEDDING_KEY,
  ROWID,
  Field,
  Item,
  MapType,
  RichData,
  field,
  lilac_embedding,
  schema,
)
from ..signal import TextEmbeddingSignal, TextSignal, clear_signal_registry, register_signal
from ..source import clear_source_registry, register_source
from .dataset import Column, DatasetManifest, config_from_dataset
from .dataset_test_utils import (
  TEST_DATASET_NAME,
  TEST_NAMESPACE,
  TestDataMaker,
  TestSource,
  enriched_item,
)

SIMPLE_ITEMS: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
]

EMBEDDINGS: list[tuple[str, list[float]]] = [
  ('hello.', [1.0, 0.0, 0.0]),
  ('hello2.', [1.0, 1.0, 0.0]),
  ('hello world.', [1.0, 1.0, 1.0]),
  ('hello world2.', [2.0, 1.0, 1.0]),
]

STR_EMBEDDINGS: dict[str, list[float]] = {text: embedding for text, embedding in EMBEDDINGS}


class TestEmbedding(TextEmbeddingSignal):
  """A test embed function."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Item]:
    """Call the embedding function."""
    for example in data:
      yield [lilac_embedding(0, len(example), np.array(STR_EMBEDDINGS[cast(str, example)]))]


class LengthSignal(TextSignal):
  name: ClassVar[str] = 'length_signal'

  _call_count: int = 0

  def fields(self) -> Field:
    return field('int32')

  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for text_content in data:
      self._call_count += 1
      yield len(text_content)


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'flen': 'float32'})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    return ({'len': len(text_content), 'flen': float(len(text_content))} for text_content in data)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_source(TestSource)
  register_signal(TestSignal)
  register_signal(TestEmbedding)
  register_signal(LengthSignal)
  register_signal(SignalWithQuoteInIt)
  register_signal(SignalWithDoubleQuoteInIt)

  # Unit test runs.
  yield

  # Teardown.
  clear_source_registry()
  clear_signal_registry()


def test_select_all_columns(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.select_rows()
  assert list(result) == SIMPLE_ITEMS


def test_select_subcols_with_dot_seperator(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'people': [{'name': 'A', 'address': {'zip': 1}}, {'name': 'B', 'address': {'zip': 2}}]},
    {'people': [{'name': 'C', 'address': {'zip': 3}}]},
  ]
  dataset = make_test_data(items)

  result = dataset.select_rows(['people.*.name', 'people.*.address.zip'])
  assert list(result) == [
    {'people.*.name': ['A', 'B'], 'people.*.address.zip': [1, 2]},
    {'people.*.name': ['C'], 'people.*.address.zip': [3]},
  ]

  result = dataset.select_rows(['people.*.address.zip'], combine_columns=True)
  assert list(result) == [
    {'people': [{'address': {'zip': 1}}, {'address': {'zip': 2}}]},
    {'people': [{'address': {'zip': 3}}]},
  ]

  result = dataset.select_rows(['people'])
  assert list(result) == items


def test_select_subcols_with_escaped_dot(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'people.new': [{'name': 'A'}, {'name': 'B'}]},
    {'people.new': [{'name': 'C'}]},
  ]
  dataset = make_test_data(items)

  result = dataset.select_rows(['"people.new".*.name'])
  assert list(result) == [{'people.new.*.name': ['A', 'B']}, {'people.new.*.name': ['C']}]

  # Escape name even though it does not need to be.
  result = dataset.select_rows(['"people.new".*."name"'])
  assert list(result) == [{'people.new.*.name': ['A', 'B']}, {'people.new.*.name': ['C']}]


def test_select_star(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [{'name': 'A', 'info': {'age': 40}}, {'name': 'B', 'info': {'age': 42}}]
  dataset = make_test_data(items)

  # Select *.
  result = dataset.select_rows(['*'])
  assert list(result) == [{'name': 'A', 'info.age': 40}, {'name': 'B', 'info.age': 42}]

  # Select (*,).
  result = dataset.select_rows([('*',)])
  assert list(result) == [{'name': 'A', 'info.age': 40}, {'name': 'B', 'info.age': 42}]

  # Select *, plus a redundant `info` column.
  result = dataset.select_rows(['*', 'info'])
  assert list(result) == [
    {'name': 'A', 'info.age': 40, 'info': {'age': 40}},
    {'name': 'B', 'info.age': 42, 'info': {'age': 42}},
  ]

  # Select * plus an inner `info.age` column.
  result = dataset.select_rows(['*', ('info', 'age')])
  assert list(result) == [
    {'info.age': 40, 'info.age_2': 40, 'name': 'A'},
    {'info.age': 42, 'info.age_2': 42, 'name': 'B'},
  ]


def test_select_star_with_combine_cols(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [{'name': 'A', 'info': {'age': 40}}, {'name': 'B', 'info': {'age': 42}}]
  dataset = make_test_data(items)

  # Select *.
  result = dataset.select_rows(['*'], combine_columns=True)
  assert list(result) == items

  # Select *, plus a redundant `info` column.
  result = dataset.select_rows(['*', 'info'], combine_columns=True)
  assert list(result) == items

  # Select * plus an inner `info.age` column.
  result = dataset.select_rows(['*', ('info', 'age')], combine_columns=True)
  assert list(result) == items

  # Select *, plus redundant `name`, plus a udf.
  udf = Column('name', signal_udf=TestSignal())
  result = dataset.select_rows(['*', 'name', udf], combine_columns=True)

  assert list(result) == [
    {'name': enriched_item('A', {'test_signal': {'len': 1, 'flen': 1.0}}), 'info': {'age': 40}},
    {'name': enriched_item('B', {'test_signal': {'len': 1, 'flen': 1.0}}), 'info': {'age': 42}},
  ]


def test_select_ids(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.select_rows([ROWID])

  assert list(result) == [{ROWID: '1'}, {ROWID: '2'}, {ROWID: '3'}]


def test_select_ids_with_limit_and_offset(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [{str(i): i} for i in range(9)]
  dataset = make_test_data(items)

  result = dataset.select_rows([ROWID], offset=1, limit=3)
  assert list(result) == [{ROWID: '2'}, {ROWID: '3'}, {ROWID: '4'}]

  result = dataset.select_rows([ROWID], offset=7, limit=2)
  assert list(result) == [{ROWID: '8'}, {ROWID: '9'}]

  result = dataset.select_rows([ROWID], offset=8, limit=200)
  assert list(result) == [{ROWID: '9'}]

  result = dataset.select_rows([ROWID], offset=9, limit=200)
  assert list(result) == []


def test_columns(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.select_rows([ROWID, 'str', 'float'])

  assert list(result) == [
    {ROWID: '1', 'str': 'a', 'float': 3.0},
    {ROWID: '2', 'str': 'b', 'float': 2.0},
    {ROWID: '3', 'str': 'b', 'float': 1.0},
  ]


def test_select_rows_iter(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.select_rows()

  def _get_rows() -> list[Item]:
    nonlocal result
    rows: list[Item] = []
    for r in result:
      rows.append(r)
    return rows

  # Make sure we can iterate twice.
  assert _get_rows() == SIMPLE_ITEMS
  assert _get_rows() == SIMPLE_ITEMS


def test_select_rows_next(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)

  result = dataset.select_rows()

  def _get_rows() -> list[Item]:
    nonlocal result
    rows: list[Item] = []
    while True:
      try:
        rows.append(next(result))
      except StopIteration:
        break

    return rows

  # Make sure we can iterate with next() twice.
  assert _get_rows() == SIMPLE_ITEMS
  assert _get_rows() == SIMPLE_ITEMS


def test_merge_values(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  test_signal = TestSignal()
  dataset.compute_signal(test_signal, 'text')
  length_signal = LengthSignal()
  dataset.compute_signal(length_signal, 'text')

  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'length_signal': 5, 'test_signal': {'len': 5, 'flen': 5.0}})},
    {
      'text': enriched_item(
        'everybody', {'length_signal': 9, 'test_signal': {'len': 9, 'flen': 9.0}}
      )
    },
  ]

  # Test subselection.
  result = dataset.select_rows(
    ['text', ('text', 'test_signal', 'flen'), ('text', 'test_signal', 'len')]
  )
  assert list(result) == [
    {'text': 'hello', 'text.test_signal.flen': 5.0, 'text.test_signal.len': 5},
    {'text': 'everybody', 'text.test_signal.flen': 9.0, 'text.test_signal.len': 9},
  ]

  # Test subselection with combine_columns=True.
  result = dataset.select_rows(
    ['text', ('text', 'test_signal', 'flen'), ('text', 'test_signal', 'len')], combine_columns=True
  )
  assert list(result) == [
    {'text': enriched_item('hello', {'length_signal': 5, 'test_signal': {'len': 5, 'flen': 5.0}})},
    {
      'text': enriched_item(
        'everybody', {'length_signal': 9, 'test_signal': {'len': 9, 'flen': 9.0}}
      )
    },
  ]

  # Test subselection with aliasing.
  result = dataset.select_rows(
    columns=['text', Column(('text', 'test_signal', 'len'), alias='metadata')]
  )
  assert list(result) == [{'text': 'hello', 'metadata': 5}, {'text': 'everybody', 'metadata': 9}]

  result = dataset.select_rows(columns=['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'length_signal': 5, 'test_signal': {'len': 5, 'flen': 5.0}})},
    {
      'text': enriched_item(
        'everybody', {'length_signal': 9, 'test_signal': {'len': 9, 'flen': 9.0}}
      )
    },
  ]


def test_enriched_select_all(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  test_signal = TestSignal()
  dataset.compute_signal(test_signal, 'text')
  length_signal = LengthSignal()
  dataset.compute_signal(length_signal, 'text')

  result = dataset.select_rows()
  assert list(result) == [
    {
      'text': 'hello',
      'text.length_signal': 5,
      'text.test_signal.len': 5,
      'text.test_signal.flen': 5.0,
    },
    {
      'text': 'everybody',
      'text.length_signal': 9,
      'text.test_signal.len': 9,
      'text.test_signal.flen': 9.0,
    },
  ]


def test_merge_array_values(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'texts': ['hello', 'everybody']}, {'texts': ['a', 'bc', 'def']}])

  test_signal = TestSignal()
  dataset.compute_signal(test_signal, ('texts', '*'))
  length_signal = LengthSignal()
  dataset.compute_signal(length_signal, ('texts', '*'))

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'texts': [
          field(
            'string',
            fields={
              'length_signal': field('int32', length_signal.model_dump(exclude_none=True)),
              'test_signal': field(
                signal=test_signal.model_dump(exclude_none=True),
                fields={'len': 'int32', 'flen': 'float32'},
              ),
            },
          )
        ]
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  result = dataset.select_rows(['texts'], combine_columns=True)
  assert list(result) == [
    {
      'texts': [
        enriched_item('hello', {'length_signal': 5, 'test_signal': {'len': 5, 'flen': 5.0}}),
        enriched_item('everybody', {'length_signal': 9, 'test_signal': {'len': 9, 'flen': 9.0}}),
      ]
    },
    {
      'texts': [
        enriched_item('a', {'length_signal': 1, 'test_signal': {'len': 1, 'flen': 1.0}}),
        enriched_item('bc', {'length_signal': 2, 'test_signal': {'len': 2, 'flen': 2.0}}),
        enriched_item('def', {'length_signal': 3, 'test_signal': {'len': 3, 'flen': 3.0}}),
      ]
    },
  ]

  # Test subselection.
  result = dataset.select_rows(
    [('texts', '*'), ('texts', '*', 'length_signal'), ('texts', '*', 'test_signal', 'flen')]
  )
  assert list(result) == [
    {
      'texts': ['hello', 'everybody'],
      'texts.*.test_signal.flen': [5.0, 9.0],
      'texts.*.length_signal': [5, 9],
    },
    {
      'texts': ['a', 'bc', 'def'],
      'texts.*.test_signal.flen': [1.0, 2.0, 3.0],
      'texts.*.length_signal': [1, 2, 3],
    },
  ]


def test_combining_columns(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'text': 'hello',
        'extra': {'text': {'length_signal': 5, 'test_signal': {'len': 5, 'flen': 5.0}}},
      },
      {
        'text': 'everybody',
        'extra': {'text': {'length_signal': 9, 'test_signal': {'len': 9, 'flen': 9.0}}},
      },
    ]
  )

  # Sub-select text and test_signal.
  result = dataset.select_rows(['text', ('extra', 'text', 'test_signal')], combine_columns=True)
  assert list(result) == [
    {'text': 'hello', 'extra': {'text': {'test_signal': {'len': 5, 'flen': 5.0}}}},
    {'text': 'everybody', 'extra': {'text': {'test_signal': {'len': 9, 'flen': 9.0}}}},
  ]

  # Sub-select text and length_signal.
  result = dataset.select_rows(['text', ('extra', 'text', 'length_signal')], combine_columns=True)
  assert list(result) == [
    {'text': 'hello', 'extra': {'text': {'length_signal': 5}}},
    {'text': 'everybody', 'extra': {'text': {'length_signal': 9}}},
  ]

  # Sub-select length_signal only.
  result = dataset.select_rows([('extra', 'text', 'length_signal')], combine_columns=True)
  assert list(result) == [
    {'extra': {'text': {'length_signal': 5}}},
    {'extra': {'text': {'length_signal': 9}}},
  ]

  # Aliases are ignored when combing columns.
  len_col = Column(('extra', 'text', 'length_signal'), alias='hello')
  result = dataset.select_rows([len_col], combine_columns=True)
  assert list(result) == [
    {'extra': {'text': {'length_signal': 5}}},
    {'extra': {'text': {'length_signal': 9}}},
  ]

  # Works with UDFs and aliases are ignored.
  udf_col = Column('text', alias='ignored', signal_udf=LengthSignal())
  result = dataset.select_rows(['text', udf_col], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {'length_signal': 5})},
    {'text': enriched_item('everybody', {'length_signal': 9})},
  ]


def test_source_joined_with_named_signal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(SIMPLE_ITEMS)
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema({'str': 'string', 'int': 'int32', 'bool': 'boolean', 'float': 'float32'}),
    num_items=3,
    source=TestSource(),
  )

  test_signal = TestSignal()
  dataset.compute_signal(test_signal, 'str')

  # Check the enriched dataset manifest has 'text' enriched.
  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'str': field(
          'string',
          fields={
            'test_signal': field(
              signal=test_signal.model_dump(exclude_none=True),
              fields={'len': 'int32', 'flen': 'float32'},
            )
          },
        ),
        'int': 'int32',
        'bool': 'boolean',
        'float': 'float32',
      }
    ),
    num_items=3,
    source=TestSource(),
  )

  result = dataset.select_rows(['str', Column(('str', 'test_signal'), alias='test_signal_on_str')])

  assert list(result) == [
    {'str': 'a', 'test_signal_on_str': {'len': 1, 'flen': 1.0}},
    {'str': 'b', 'test_signal_on_str': {'len': 1, 'flen': 1.0}},
    {'str': 'b', 'test_signal_on_str': {'len': 1, 'flen': 1.0}},
  ]


def test_invalid_column_paths(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'text': enriched_item('hello', {'test_signal': {'len': 5}}),
        'text2': [
          enriched_item('hello', {'test_signal': {'len': 5}}),
          enriched_item('hi', {'test_signal': {'len': 2}}),
        ],
      }
    ]
  )

  with pytest.raises(ValueError, match='Path part "invalid" not found in the dataset'):
    dataset.select_rows([('text', 'test_signal', 'invalid')])

  with pytest.raises(ValueError, match='Selecting a specific index of a repeated field'):
    dataset.select_rows([('text2', '4', 'test_signal')])


def test_signal_with_quote(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'world'}])
  dataset.compute_signal(SignalWithQuoteInIt(), 'text')
  dataset.compute_signal(SignalWithDoubleQuoteInIt(), 'text')
  result = dataset.select_rows(['text'], combine_columns=True)
  assert list(result) == [
    {'text': enriched_item('hello', {"test'signal": True, 'test"signal': True})},
    {'text': enriched_item('world', {"test'signal": True, 'test"signal': True})},
  ]

  result = dataset.select_rows(
    ['text', "text.test'signal", 'text.test"signal'], combine_columns=False
  )
  assert list(result) == [
    {'text': 'hello', "text.test'signal": True, 'text.test"signal': True},
    {'text': 'world', "text.test'signal": True, 'text.test"signal': True},
  ]


class SignalWithQuoteInIt(TextSignal):
  name: ClassVar[str] = "test'signal"

  @override
  def fields(self) -> Field:
    return field('boolean')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for d in data:
      yield True


class SignalWithDoubleQuoteInIt(TextSignal):
  name: ClassVar[str] = 'test"signal'

  @override
  def fields(self) -> Field:
    return field('boolean')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    for d in data:
      yield True


def test_config_from_dataset(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello world.'}])
  dataset.compute_signal(TestSignal(), 'text')
  dataset.compute_embedding('test_embedding', 'text')

  assert dataset.manifest() == DatasetManifest(
    namespace=TEST_NAMESPACE,
    dataset_name=TEST_DATASET_NAME,
    data_schema=schema(
      {
        'text': field(
          'string',
          fields={
            'test_signal': field(
              signal=TestSignal().model_dump(exclude_none=True),
              fields={'len': 'int32', 'flen': 'float32'},
            ),
            'test_embedding': field(
              signal=TestEmbedding().model_dump(exclude_none=True),
              fields=[field('string_span', fields={EMBEDDING_KEY: 'embedding'})],
            ),
          },
        )
      }
    ),
    num_items=2,
    source=TestSource(),
  )

  config = config_from_dataset(dataset)

  assert config == DatasetConfig(
    namespace=TEST_NAMESPACE,
    name=TEST_DATASET_NAME,
    source=TestSource(),
    embeddings=[EmbeddingConfig(path=('text',), embedding='test_embedding')],
    signals=[SignalConfig(path=('text',), signal=TestSignal())],
    settings=DatasetSettings(ui=DatasetUISettings(media_paths=[('text',)])),
  )


def test_map_dtype(make_test_data: TestDataMaker) -> None:
  items = [
    {'column': {'a': 1.0, 'b': 2.0}},
    {'column': {'b': 2.5}},
    {'column': {'a': 3.0, 'c': 3.5}},
  ]
  map_dtype = MapType(key_type='string', value_field=field('float32'))
  data_schema = schema({'column': Field(dtype=map_dtype)})
  dataset = make_test_data(items, schema=data_schema)
  assert dataset.manifest() == DatasetManifest(
    namespace='test_namespace',
    dataset_name='test_dataset',
    data_schema=schema(
      {
        'column': Field(
          dtype=map_dtype,
          fields={'a': field('float32'), 'b': field('float32'), 'c': field('float32')},
        )
      }
    ),
    source=TestSource(),
    num_items=3,
  )

  result = dataset.select_rows(['column'])
  assert list(result) == [
    {'column': {'key': ['a', 'b'], 'value': [1.0, 2.0]}},
    {'column': {'key': ['b'], 'value': [2.5]}},
    {'column': {'key': ['a', 'c'], 'value': [3.0, 3.5]}},
  ]


def test_get_embeddings(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello.'}, {'text': 'hello world.'}])
  dataset.compute_embedding('test_embedding', 'text')

  rows = list(dataset.select_rows([ROWID]))

  span_vectors = dataset.get_embeddings('test_embedding', rows[0][ROWID], 'text')
  assert len(span_vectors) == 1
  assert span_vectors[0]['span'] == (0, 6)
  np.testing.assert_array_equal(
    span_vectors[0]['vector'], np.array([1.0, 0.0, 0.0], dtype=np.float32)
  )

  span_vectors = dataset.get_embeddings('test_embedding', rows[1][ROWID], 'text')
  assert len(span_vectors) == 1
  assert span_vectors[0]['span'] == (0, 12)
  np.testing.assert_array_equal(
    span_vectors[0]['vector'], np.array([1.0, 1.0, 1.0], dtype=np.float32)
  )



"""Test the Substring Search signal."""

from typing import cast

import pytest
from pydantic import ValidationError

from ..schema import field
from ..splitters.text_splitter_test_utils import text_to_expected_spans
from .substring_search import SubstringSignal


def test_substring_fields() -> None:
  signal = SubstringSignal(query='test')
  assert signal.fields() == field(fields=['string_span'])


def test_query_is_required() -> None:
  with pytest.raises(ValidationError):
    SubstringSignal(query=cast(str, None))


def test_compute() -> None:
  signal = SubstringSignal(query='test')

  text = 'The word TEST shows up 3 times, teST and test'
  spans = list(signal.compute([text]))

  expected_spans = text_to_expected_spans(text, ['TEST', 'teST', 'test'])
  assert [expected_spans] == spans



"""Test signal base class."""
from typing import ClassVar, Iterable, Iterator, Optional

import pytest
from typing_extensions import override

from .schema import Field, Item, RichData, SignalInputType, field
from .signal import (
  Signal,
  TextEmbeddingSignal,
  clear_signal_registry,
  get_signal_by_type,
  get_signal_cls,
  get_signals_by_type,
  register_signal,
  resolve_signal,
)


class TestSignal(Signal):
  """A test signal."""

  # Pydantic fields
  name: ClassVar[str] = 'test_signal'
  input_type: ClassVar[SignalInputType] = SignalInputType.TEXT

  query: str

  @override
  def fields(self) -> Field:
    return field('float32')

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    del data
    yield from ()


class TestTextEmbedding(TextEmbeddingSignal):
  """A test text embedding."""

  name: ClassVar[str] = 'test_embedding'

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    del data
    yield from ()


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  register_signal(TestSignal)
  register_signal(TestTextEmbedding)

  # Unit test runs.
  yield

  # Teardown.
  clear_signal_registry()


def test_signal_serialization() -> None:
  signal = TestSignal(query='test')

  # The class variables should not be included.
  assert signal.model_dump(exclude_none=True) == {
    'signal_name': 'test_signal',
    'query': 'test',
    'remote': False,
  }


def test_get_signal_cls() -> None:
  """Test getting a signal."""
  assert TestSignal == get_signal_cls('test_signal')


def test_resolve_signal() -> None:
  """Test resolving a signal."""
  test_signal = TestSignal(query='hello')

  # Signals pass through.
  assert resolve_signal(test_signal) == test_signal

  # Dicts resolve to the base class.
  assert resolve_signal(test_signal.model_dump()) == test_signal


def test_get_signal_by_type() -> None:
  assert get_signal_by_type(TestTextEmbedding.name, TextEmbeddingSignal) == TestTextEmbedding


def test_get_signal_by_type_validation() -> None:
  with pytest.raises(ValueError, match='Signal "invalid_signal" not found in the registry'):
    get_signal_by_type('invalid_signal', TestTextEmbedding)

  with pytest.raises(ValueError, match=f'"{TestSignal.name}" is a `{TestSignal.__name__}`'):
    get_signal_by_type(TestSignal.name, TextEmbeddingSignal)


def test_get_signals_by_type() -> None:
  assert get_signals_by_type(TextEmbeddingSignal) == [TestTextEmbedding]


class TestSignalNoDisplayName(Signal):
  name: ClassVar[str] = 'signal_no_name'


class TestSignalDisplayName(Signal):
  name: ClassVar[str] = 'signal_display_name'
  display_name: ClassVar[str] = 'test display name'


def test_signal_title_schema() -> None:
  assert TestSignalNoDisplayName.model_json_schema()['title'] == TestSignalNoDisplayName.__name__
  assert TestSignalDisplayName.model_json_schema()['title'] == 'test display name'



"""Utilities for unit tests."""
import pathlib
import uuid
from datetime import datetime
from typing import Type, cast

import pyarrow.parquet as pq
from pydantic import BaseModel

from .schema import ROWID, Item
from .source import SourceManifest

TEST_TIME = datetime(2023, 8, 15, 1, 23, 45)


def retrieve_parquet_rows(
  tmp_path: pathlib.Path, manifest: SourceManifest, retain_rowid: bool = False
) -> list[Item]:
  """Retrieve the rows from a parquet source."""
  items = []
  for file in manifest.files:
    items.extend(pq.read_table(tmp_path / file).to_pylist())
  for item in items:
    assert ROWID in item
    if not retain_rowid:
      del item[ROWID]
  return items


def allow_any_datetime(cls: Type[BaseModel]) -> None:
  """Overrides the __eq__ method for a pydantic model to allow any datetimes to be equal."""

  # freeze_time and Dask don't play well together so we override equality here.
  def _replace_datetimes(self: dict) -> None:
    # Recursively replace datetimes in both self and other with TEST_TIME.
    for key in self:
      if isinstance(self[key], dict):
        _replace_datetimes(self[key])
      elif isinstance(self[key], datetime):
        self[key] = TEST_TIME

  def _new_eq(self: BaseModel, other: object) -> bool:
    self_dict = self.model_dump()
    other_dict = cast(BaseModel, other).model_dump()
    _replace_datetimes(self_dict)
    return self_dict == other_dict

  cls.__eq__ = _new_eq  # type: ignore


def fake_uuid(id: bytes) -> uuid.UUID:
  """Create a test UUID."""
  return uuid.UUID((id * 16).hex())



"""Tests for the pandas source."""

import pandas as pd

from ..schema import schema
from ..source import SourceSchema
from .pandas_source import PANDAS_INDEX_COLUMN, PandasSource


def test_simple_dataframe() -> None:
  df = pd.DataFrame.from_records(
    [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}]
  )

  source = PandasSource(df)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({PANDAS_INDEX_COLUMN: 'int64', 'name': 'string', 'age': 'int64'}).fields,
    num_items=3,
  )

  items = list(source.yield_items())

  assert items == [
    {PANDAS_INDEX_COLUMN: 0, 'name': 'a', 'age': 1},
    {PANDAS_INDEX_COLUMN: 1, 'name': 'b', 'age': 2},
    {PANDAS_INDEX_COLUMN: 2, 'name': 'c', 'age': 3},
  ]


def test_simple_dataframe_with_index() -> None:
  df = pd.DataFrame.from_records(
    [{'name': 'a', 'age': 1}, {'name': 'b', 'age': 2}, {'name': 'c', 'age': 3}],
    index=['id1', 'id2', 'id3'],
  )

  source = PandasSource(df)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({PANDAS_INDEX_COLUMN: 'string', 'name': 'string', 'age': 'int64'}).fields,
    num_items=3,
  )

  items = list(source.yield_items())

  # The PANDAS_INDEX_COLUMN aligns with the pandas index.
  assert items == [
    {PANDAS_INDEX_COLUMN: 'id1', 'name': 'a', 'age': 1},
    {PANDAS_INDEX_COLUMN: 'id2', 'name': 'b', 'age': 2},
    {PANDAS_INDEX_COLUMN: 'id3', 'name': 'c', 'age': 3},
  ]



"""Clustering utilities."""
import functools
import gc
import itertools
import random
from typing import Any, Callable, Iterator, Optional, Union, cast

import instructor
import modal
import numpy as np
from instructor.exceptions import IncompleteOutputException
from joblib import Parallel, delayed
from pydantic import (
  BaseModel,
)
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_random_exponential

from ..batch_utils import compress_docs, flatten_path_iter, group_by_sorted_key_iter
from ..dataset_format import DatasetFormatInputSelector
from ..embeddings.jina import JinaV2Small
from ..schema import (
  EMBEDDING_KEY,
  PATH_WILDCARD,
  VALUE_KEY,
  ClusterInfo,
  ClusterInputFormatSelectorInfo,
  Item,
  Path,
  PathTuple,
  field,
  normalize_path,
)
from ..signal import (
  TopicFn,
)
from ..tasks import TaskId, TaskInfo, get_task_manager
from ..utils import DebugTimer, chunks
from .dataset import Dataset
from .dataset_utils import (
  get_callable_name,
  sparse_to_dense_compute,
)

_SHORTEN_LEN = 400
_TOP_K_CENTRAL_DOCS = 7
_TOP_K_CENTRAL_TITLES = 15
_NUM_THREADS = 32

CLUSTER_ID = 'cluster_id'
CLUSTER_MEMBERSHIP_PROB = 'cluster_membership_prob'
CLUSTER_TITLE = 'cluster_title'

CATEGORY_ID = 'category_id'
CATEGORY_MEMBERSHIP_PROB = 'category_membership_prob'
CATEGORY_TITLE = 'category_title'

FIELD_SUFFIX = 'cluster'

MIN_CLUSTER_SIZE = 5
UMAP_DIM = 5
UMAP_SEED = 42
HDBSCAN_SELECTION_EPS = 0.05
BATCH_SOFT_CLUSTER_NOISE = 1024


@functools.cache
def _openai_client() -> Any:
  """Get an OpenAI client."""
  try:
    import openai

  except ImportError:
    raise ImportError(
      'Could not import the "openai" python package. '
      'Please install it with `pip install openai`.'
    )

  return instructor.patch(openai.OpenAI())


def _snippet_to_prefix_and_suffix(text: str) -> str:
  text = text.strip()
  if len(text) <= _SHORTEN_LEN:
    return text
  prefix_len = _SHORTEN_LEN // 2
  return text[:prefix_len] + '\n...\n' + text[-prefix_len:]


class Title(BaseModel):
  """A 4-5 word title for the group of related requests."""

  title: str


def summarize_request(ranked_docs: list[tuple[str, float]]) -> str:
  """Summarize a group of requests in a title of at most 5 words."""
  # Get the top 5 documents.
  docs = [doc for doc, _ in ranked_docs[:_TOP_K_CENTRAL_DOCS]]
  texts = [f'BEGIN_REQUEST\n{_snippet_to_prefix_and_suffix(doc)}\nEND_REQUEST' for doc in docs]
  input = '\n'.join(texts)
  try:
    import openai

  except ImportError:
    raise ImportError(
      'Could not import the "openai" python package. '
      'Please install it with `pip install openai`.'
    )

  @retry(
    retry=retry_if_exception_type((openai.RateLimitError, openai.APITimeoutError)),
    wait=wait_random_exponential(multiplier=0.5, max=60),
    stop=stop_after_attempt(10),
  )
  def request_with_retries() -> str:
    max_tokens = 50
    while max_tokens <= 200:
      try:
        title = _openai_client().chat.completions.create(
          model='gpt-3.5-turbo-1106',
          response_model=Title,
          temperature=0.0,
          max_tokens=max_tokens,
          messages=[
            {
              'role': 'system',
              'content': (
                'You are a world-class title generator. Ignore the group of related requests '
                'below, and generate a short title to describe the common theme. Some examples: '
                '"YA book reviews", "Questions about South East Asia", "Translating English to '
                'Polish", "Writing product descriptions", etc. Prefer using descriptive words. Do '
                'not use vague words like "various", "assortment", "comments", "discussion", etc.'
              ),
            },
            {'role': 'user', 'content': input},
          ],
        )
        return title.title
      except IncompleteOutputException:
        max_tokens = max_tokens * 2
        print(f'Retrying with max_tokens={max_tokens}')
    print(f'Could not generate a reasonable title for input:\n{input}')
    return 'FAILED_TO_GENERATE'

  return request_with_retries()


class Category(BaseModel):
  """A short category title."""

  category: str


def generate_category(ranked_docs: list[tuple[str, float]]) -> str:
  """Summarize a list of titles in a category."""
  # Get the top 5 documents.
  docs = [doc for doc, _ in ranked_docs[:_TOP_K_CENTRAL_TITLES]]
  input = '\n'.join(docs)
  try:
    import openai

  except ImportError:
    raise ImportError(
      'Could not import the "openai" python package. '
      'Please install it with `pip install openai`.'
    )

  @retry(
    retry=retry_if_exception_type((openai.RateLimitError, openai.APITimeoutError)),
    wait=wait_random_exponential(multiplier=0.5, max=60),
    stop=stop_after_attempt(10),
  )
  def request_with_retries() -> str:
    category = _openai_client().chat.completions.create(
      model='gpt-3.5-turbo-1106',
      response_model=Category,
      temperature=0.0,
      max_tokens=50,
      messages=[
        {
          'role': 'system',
          'content': (
            'You are a world-class category labeler. Generate a short category name for the '
            'provided titles. For example, given two titles "translating english to polish" and '
            '"translating korean to english", generate "Translation".'
          ),
        },
        {'role': 'user', 'content': input},
      ],
    )
    return category.category

  return request_with_retries()


def _compute_titles(
  items: Iterator[Item],
  text_column: str,
  cluster_id_column: str,
  membership_column: str,
  topic_fn: TopicFn,
  task_info: Optional[TaskInfo] = None,
) -> Iterator[str]:
  def _compute_title(
    sorted_docs: list[tuple[str, float]], group_size: int
  ) -> Optional[tuple[int, Optional[str]]]:
    if not sorted_docs:
      return group_size, None
    return group_size, topic_fn(sorted_docs)

  def _delayed_compute_all_titles() -> Iterator:
    for group in group_by_sorted_key_iter(items, lambda x: x[cluster_id_column]):
      sorted_docs: list[tuple[str, float]] = []

      for item in group:
        if not item:
          continue

        cluster_id = item.get(cluster_id_column, -1)
        if cluster_id < 0:
          continue

        text = item.get(text_column)
        if not text:
          continue

        membership_prob = item.get(membership_column, 0)
        if membership_prob == 0:
          continue

        sorted_docs.append((text, membership_prob))

      # Remove any duplicate texts in the group.
      sorted_docs = list(set(sorted_docs))

      # Shuffle the group to avoid biasing the topic function.
      random.shuffle(sorted_docs)

      # Sort the group by membership probability after shuffling so that we still choose high
      # membership scores but they are still shuffled when the values are equal.
      sorted_docs.sort(key=lambda text_score: text_score[1], reverse=True)

      yield delayed(_compute_title)(sorted_docs, len(group))

  parallel = Parallel(n_jobs=_NUM_THREADS, backend='threading', return_as='generator')
  if task_info:
    task_info.total_progress = 0
  for group_size, title in parallel(_delayed_compute_all_titles()):
    if task_info:
      task_info.total_progress += group_size
    for _ in range(group_size):
      yield title


def cluster_impl(
  dataset: Dataset,
  input_fn_or_path: Union[Path, Callable[[Item], str], DatasetFormatInputSelector],
  output_path: Optional[Path] = None,
  min_cluster_size: int = 5,
  topic_fn: TopicFn = summarize_request,
  overwrite: bool = False,
  remote: bool = False,
  task_id: Optional[TaskId] = None,
  recompute_titles: bool = False,
) -> None:
  """Compute clusters for a field of the dataset."""
  task_manager = get_task_manager()
  task_info: Optional[TaskInfo] = None
  if task_id:
    task_info = task_manager.get_task_info(task_id)
  manifest = dataset.manifest()
  schema = manifest.data_schema
  path: Optional[PathTuple] = None

  dataset_format_input_selector: Optional[DatasetFormatInputSelector] = None
  if isinstance(input_fn_or_path, DatasetFormatInputSelector):
    dataset_format_input_selector = input_fn_or_path
    input_fn_or_path = input_fn_or_path.selector

  if not callable(input_fn_or_path):
    path = normalize_path(input_fn_or_path)
    # Make sure the path exists.
    if not schema.has_field(path):
      raise ValueError(f'Path {path} does not exist in the dataset.')
    input_field = schema.get_field(path)
    if not input_field.dtype or input_field.dtype.type != 'string':
      raise ValueError(f'Path {path} must be a string field.')

  elif not output_path:
    raise ValueError('output_path must be provided if input is a function.')

  # Output the cluster enrichment to a sibling path, unless an output path is provided by the user.
  if output_path:
    cluster_output_path = normalize_path(output_path)
  elif path:
    # The sibling output path is the same as the input path, but with a different suffix.
    index = 0
    for i, path_part in enumerate(path):
      if path_part == PATH_WILDCARD:
        break
      else:
        index = i

    parent = path[:index]
    sibling = '_'.join([p for p in path[index:] if p != PATH_WILDCARD])
    cluster_output_path = (*parent, f'{sibling}__{FIELD_SUFFIX}')
  else:
    raise ValueError('input must be provided.')

  # Extract the text from the input path into a temporary column.
  TEXT_COLUMN = 'text'
  temp_text_path = (*cluster_output_path, TEXT_COLUMN)
  temp_path_exists = schema.has_field(temp_text_path)
  if not temp_path_exists or overwrite:
    # Since input is a function, map over the dataset to make a temporary column with that text.
    if task_info:
      task_info.message = 'Extracting text from items'

    def _flatten_input(item: Item, input_path: PathTuple) -> str:
      texts = flatten_path_iter(item, input_path)
      # Filter out Nones
      texts = (t for t in texts if t)
      # Deal with enriched items.
      texts = (t[VALUE_KEY] if VALUE_KEY in t else t for t in texts)
      return '\n'.join(texts)

    def extract_text(item: Item) -> Item:
      cluster_item = item
      for path_part in cluster_output_path:
        cluster_item = cluster_item.get(path_part, {})

      text = (
        input_fn_or_path(item)
        if callable(input_fn_or_path)
        else _flatten_input(item, cast(PathTuple, path))
      )
      return {**cluster_item, TEXT_COLUMN: text}

    dataset.map(extract_text, output_path=cluster_output_path, overwrite=True)

  cluster_ids_exists = schema.has_field((*cluster_output_path, CLUSTER_ID))
  if not cluster_ids_exists or overwrite:
    if task_info:
      task_info.message = 'Computing clusters'
      task_info.total_progress = 0
      task_info.total_len = None

    def compute_clusters(items: Iterator[Item]) -> Iterator[Item]:
      items, items2 = itertools.tee(items)
      docs: Iterator[Optional[str]] = (item.get(TEXT_COLUMN) for item in items)
      cluster_items = sparse_to_dense_compute(
        docs, lambda x: _hdbscan_cluster(x, min_cluster_size, remote)
      )
      for item, cluster_item in zip(items2, cluster_items):
        yield {**item, **(cluster_item or {})}

    # Compute the clusters.
    dataset.transform(
      compute_clusters,
      input_path=cluster_output_path,
      output_path=cluster_output_path,
      overwrite=True,
    )

  total_len = dataset.stats(temp_text_path).total_count
  cluster_titles_exist = schema.has_field((*cluster_output_path, CLUSTER_TITLE))
  if not cluster_titles_exist or overwrite or recompute_titles:
    if task_info:
      task_info.message = 'Computing cluster titles'
      task_info.total_progress = 0
      task_info.total_len = total_len

    def compute_cluster_titles(items: Iterator[Item]) -> Iterator[Item]:
      items, items2 = itertools.tee(items)
      titles = _compute_titles(
        items,
        text_column=TEXT_COLUMN,
        cluster_id_column=CLUSTER_ID,
        membership_column=CLUSTER_MEMBERSHIP_PROB,
        topic_fn=topic_fn,
        task_info=task_info,
      )
      for item, title in zip(items2, titles):
        yield {**item, CLUSTER_TITLE: title}

    dataset.transform(
      compute_cluster_titles,
      input_path=cluster_output_path,
      output_path=cluster_output_path,
      sort_by=(*cluster_output_path, CLUSTER_ID),
      overwrite=True,
    )

  category_id_exists = schema.has_field((*cluster_output_path, CATEGORY_ID))
  if not category_id_exists or overwrite or recompute_titles:
    if task_info:
      task_info.message = 'Computing super clusters'
      task_info.total_progress = 0
      task_info.total_len = None

    def compute_category_clusters(items: Iterator[Item]) -> Iterator[Item]:
      items, items2 = itertools.tee(items)
      docs = (item.get(CLUSTER_TITLE) for item in items)
      cluster_items = sparse_to_dense_compute(
        docs, lambda x: _hdbscan_cluster(x, min_cluster_size, remote)
      )
      for item, cluster_item in zip(items2, cluster_items):
        item[CATEGORY_ID] = (cluster_item or {}).get(CLUSTER_ID, -1)
        item[CATEGORY_MEMBERSHIP_PROB] = (cluster_item or {}).get(CLUSTER_MEMBERSHIP_PROB, 0)
        yield item

    # Compute the clusters.
    dataset.transform(
      compute_category_clusters,
      input_path=cluster_output_path,
      output_path=cluster_output_path,
      overwrite=True,
    )

  category_title_path = (*cluster_output_path, CATEGORY_TITLE)
  category_title_exists = schema.has_field(category_title_path)
  if not category_title_exists or overwrite or recompute_titles:
    if task_info:
      task_info.message = 'Computing category titles'
      task_info.total_progress = 0
      task_info.total_len = total_len

    def compute_category_titles(items: Iterator[Item]) -> Iterator[Item]:
      items, items2 = itertools.tee(items)
      titles = _compute_titles(
        items,
        text_column=CLUSTER_TITLE,
        cluster_id_column=CATEGORY_ID,
        membership_column=CATEGORY_MEMBERSHIP_PROB,
        topic_fn=generate_category,
        task_info=task_info,
      )
      for item, title in zip(items2, titles):
        # Drop the temporary newline-concatenated text column.
        del item[TEXT_COLUMN]
        yield {**item, CATEGORY_TITLE: title}

    dataset.transform(
      compute_category_titles,
      input_path=cluster_output_path,
      output_path=cluster_output_path,
      sort_by=(*cluster_output_path, CATEGORY_ID),
      overwrite=True,
      schema=field(
        fields={
          CLUSTER_ID: field('int32', categorical=True),
          CLUSTER_MEMBERSHIP_PROB: 'float32',
          CLUSTER_TITLE: 'string',
          CATEGORY_ID: field('int32', categorical=True),
          CATEGORY_MEMBERSHIP_PROB: 'float32',
          CATEGORY_TITLE: 'string',
        },
        cluster=ClusterInfo(
          min_cluster_size=min_cluster_size,
          remote=remote,
          input_path=(get_callable_name(input_fn_or_path),) if callable(input_fn_or_path) else path,
          input_format_selector=ClusterInputFormatSelectorInfo(
            format=manifest.dataset_format.name,
            selector=dataset_format_input_selector.name,
          )
          if dataset_format_input_selector and manifest.dataset_format
          else None,
        ),
      ),
    )

  if task_id:
    task_manager.set_completed(task_id)


def _hdbscan_cluster(
  docs: Iterator[str],
  min_cluster_size: int = MIN_CLUSTER_SIZE,
  remote: bool = False,
) -> Iterator[Item]:
  """Cluster docs with HDBSCAN."""
  if remote:
    remote_fn = modal.Function.lookup('cluster', 'Cluster.cluster').remote
    gzipped_docs = compress_docs(list(docs))
    response = remote_fn({'gzipped_docs': gzipped_docs})
    yield from response['clusters']

  with DebugTimer('Computing embeddings'):
    jina = JinaV2Small()
    jina.setup()
    response = jina.compute(list(docs))
    jina.teardown()

  all_vectors = np.array([r[0][EMBEDDING_KEY] for r in response], dtype=np.float32)
  del response, docs
  gc.collect()

  # Use UMAP to reduce the dimensionality before hdbscan to speed up clustering.
  # For details on hyperparameters, see:
  # https://umap-learn.readthedocs.io/en/latest/clustering.html

  # Try to import the cuml version of UMAP, which is much faster than the sklearn version.
  # if CUDA is available.
  try:
    from cuml import UMAP  # type: ignore
  except ImportError:
    from umap import UMAP

  dim = all_vectors[0].size
  with DebugTimer(f'UMAP: Reducing dim from {dim} to {UMAP_DIM} of {len(all_vectors)} vectors'):
    n_neighbors = min(30, len(all_vectors) - 1)
    if UMAP_DIM < dim and UMAP_DIM < len(all_vectors):
      reducer = UMAP(
        n_components=UMAP_DIM,
        n_neighbors=n_neighbors,
        min_dist=0.0,
        n_jobs=-1,
        random_state=UMAP_SEED,
      )
      all_vectors = reducer.fit_transform(all_vectors)

  gc.collect()

  # Try to import the cuml version of HDBSCAN, which is much faster than the sklearn version.
  # if CUDA is available.
  try:
    from cuml.cluster.hdbscan import HDBSCAN, membership_vector  # type: ignore
  except ImportError:
    from hdbscan import HDBSCAN, membership_vector

  with DebugTimer('HDBSCAN: Clustering'):
    min_cluster_size = min(min_cluster_size, len(all_vectors))
    clusterer = HDBSCAN(
      min_cluster_size=min_cluster_size,
      min_samples=min_cluster_size - 1,
      cluster_selection_epsilon=HDBSCAN_SELECTION_EPS,
      cluster_selection_method='leaf',
      prediction_data=True,
    )
    clusterer.fit(all_vectors)

  noisy_vectors: list[np.ndarray] = []
  for i, cluster_id in enumerate(clusterer.labels_):
    if cluster_id == -1:
      noisy_vectors.append(all_vectors[i])
  num_noisy = len(noisy_vectors)
  perc_noisy = 100 * num_noisy / len(clusterer.labels_)
  print(f'{num_noisy} noise points ({perc_noisy:.1f}%) will be assigned to nearest cluster.')

  noisy_labels: list[np.ndarray] = []
  noisy_probs: list[np.ndarray] = []
  labels = clusterer.labels_
  memberships = clusterer.probabilities_
  if num_noisy > 0 and num_noisy < len(clusterer.labels_):
    with DebugTimer('HDBSCAN: Computing membership for the noise points'):
      for batch_noisy_vectors in chunks(noisy_vectors, BATCH_SOFT_CLUSTER_NOISE):
        batch_noisy_vectors = np.array(batch_noisy_vectors, dtype=np.float32)
        soft_clusters = membership_vector(clusterer, batch_noisy_vectors)
        if soft_clusters.ndim < 2:
          soft_clusters = soft_clusters.reshape(-1, 1)
        noisy_labels.append(np.argmax(soft_clusters, axis=1))
        noisy_probs.append(np.max(soft_clusters, axis=1))

    noisy_labels = np.concatenate(noisy_labels, axis=0, dtype=np.int32)
    noisy_probs = np.concatenate(noisy_probs, axis=0, dtype=np.float32)
    noise_index = 0
    for i, cluster_id in enumerate(labels):
      if cluster_id == -1:
        labels[i] = noisy_labels[noise_index]
        memberships[i] = noisy_probs[noise_index]
        noise_index += 1

  del clusterer, all_vectors, noisy_vectors
  gc.collect()

  for cluster_id, membership_prob in zip(labels, memberships):
    yield {CLUSTER_ID: int(cluster_id), CLUSTER_MEMBERSHIP_PROB: float(membership_prob)}



"""Router for RAG."""


from typing import Sequence, Union

from fastapi import APIRouter
from pydantic import BaseModel

from .db_manager import get_dataset
from .rag.rag import (
  RagGenerationResult,
  RagRetrievalResultItem,
  get_rag_generation,
  get_rag_retrieval,
)
from .router_dataset import Column, Filter
from .router_utils import RouteErrorHandler
from .schema import Path

router = APIRouter(route_class=RouteErrorHandler)


class RagRetrievalOptions(BaseModel):
  """The config for the rag retrieval."""

  dataset_namespace: str
  dataset_name: str
  embedding: str

  query: str

  # The main column that will be used for the retrieval. This should have an embedding index already
  # computed.
  path: Path
  # Columns that will be used for additional metadata information.
  metadata_columns: Sequence[Union[Column, Path]] = []
  filters: Sequence[Filter] = []

  # Hyper-parameters.
  chunk_window: int = 1
  top_k: int = 10
  similarity_threshold: float = 0.0


class RagGenerationOptions(BaseModel):
  """The config for the rag generation."""

  query: str
  retrieval_results: list[RagRetrievalResultItem]
  prompt_template: str


@router.post('/retrieval')
def retrieval(options: RagRetrievalOptions) -> list[RagRetrievalResultItem]:
  """Get the retrieval results for a prompt."""
  ds = get_dataset(options.dataset_namespace, options.dataset_name)
  return get_rag_retrieval(
    path=options.path,
    dataset=ds,
    embedding=options.embedding,
    query=options.query,
    top_k=options.top_k,
    chunk_window=options.chunk_window,
  )


@router.post('/generate')
def generate(options: RagGenerationOptions) -> RagGenerationResult:
  """Get the retrieval results for a prompt."""
  return get_rag_generation(
    query=options.query,
    retrieval_results=options.retrieval_results,
    prompt_template=options.prompt_template,
  )



"""Tests for inferring the sharegpt format."""


from ..data.dataset_test_utils import TestDataMaker
from .sharegpt import ShareGPT


def test_infer_sharegpt(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'conversations': [
          {'from': 'system', 'value': 'You are a language model.'},
          {'from': 'human', 'value': 'hello'},
        ]
      },
      {'conversations': [{'from': 'human', 'value': 'Hello again'}]},
    ]
  )

  assert dataset.manifest().dataset_format == ShareGPT()


def test_infer_sharegpt_extra(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'conversations': [
          {'from': 'system', 'value': 'You are a language model.'},
          {'from': 'human', 'value': 'hello'},
        ],
        'extra': 2,
      },
      {'conversations': [{'from': 'human', 'value': 'Hello again'}], 'extra': 1},
    ]
  )

  assert dataset.manifest().dataset_format == ShareGPT()


def test_human_selector(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {
        'conversations': [
          {'from': 'system', 'value': 'You are a language model.'},
          {'from': 'human', 'value': 'hello'},
          {'from': 'gpt', 'value': 'hello'},
        ]
      },
      {
        'conversations': [
          {'from': 'human', 'value': 'Hello again'},
          {'from': 'gpt', 'value': 'Hello'},
        ]
      },
      {'conversations': [{'from': 'gpt', 'value': 'i am an ai'}]},
    ]
  )

  assert list(dataset.map(map_fn=ShareGPT.system.selector)) == ['You are a language model.', '', '']
  assert list(dataset.map(map_fn=ShareGPT.human.selector)) == ['hello', 'Hello again', '']
  assert list(dataset.map(map_fn=ShareGPT.gpt.selector)) == ['hello', 'Hello', 'i am an ai']

  assert list(dataset.map(map_fn=ShareGPT.input_selectors['system'].selector)) == list(
    dataset.map(map_fn=ShareGPT.system.selector)
  )
  assert list(dataset.map(map_fn=ShareGPT.input_selectors['human'].selector)) == list(
    dataset.map(map_fn=ShareGPT.human.selector)
  )
  assert list(dataset.map(map_fn=ShareGPT.input_selectors['gpt'].selector)) == list(
    dataset.map(map_fn=ShareGPT.gpt.selector)
  )



"""Implementation-agnostic tests for exporting a dataset."""

import csv
import json
import pathlib
from datetime import datetime
from typing import ClassVar, Iterable, Iterator, Optional

import numpy as np
import pandas as pd
import pytest
from freezegun import freeze_time
from typing_extensions import override

from ..schema import ROWID, Field, Item, RichData, field
from ..signal import TextSignal, clear_signal_registry, register_signal
from .dataset import DELETED_LABEL_NAME
from .dataset_test_utils import TestDataMaker


class TestSignal(TextSignal):
  name: ClassVar[str] = 'test_signal'

  @override
  def fields(self) -> Field:
    return field(fields={'len': 'int32', 'flen': 'float32'})

  @override
  def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
    return ({'len': len(text_content), 'flen': float(len(text_content))} for text_content in data)


@pytest.fixture(scope='module', autouse=True)
def setup_teardown() -> Iterable[None]:
  # Setup.
  clear_signal_registry()
  register_signal(TestSignal)

  yield  # Unit test runs.
  clear_signal_registry()  # Teardown.


def test_export_to_json(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  dataset.compute_signal(TestSignal(), 'text')

  # Download all columns.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath)

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [
    {'text': 'hello', 'text.test_signal.len': 5, 'text.test_signal.flen': 5.0},
    {'text': 'everybody', 'text.test_signal.len': 9, 'text.test_signal.flen': 9.0},
  ]

  # Download a subset of columns with filter.
  filepath = tmp_path / 'dataset2.json'
  dataset.to_json(
    filepath,
    columns=['text', 'text.test_signal.flen'],
    filters=[('text.test_signal.len', 'greater', '6')],
  )

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [{'text': 'everybody', 'text.test_signal.flen': 9.0}]

  filepath = tmp_path / 'dataset3.json'
  dataset.to_json(filepath, filters=[('text.test_signal.flen', 'less_equal', '5')])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [
    {'text': 'hello', 'text.test_signal.len': 5, 'text.test_signal.flen': 5.0}
  ]


def test_export_to_csv(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  dataset.compute_signal(TestSignal(), 'text')

  # Download all columns.
  filepath = tmp_path / 'dataset.csv'
  dataset.to_csv(filepath)

  with open(filepath) as f:
    rows = list(csv.reader(f))

  assert rows == [
    ['text', 'text.test_signal.len', 'text.test_signal.flen'],
    ['hello', '5', '5.0'],
    ['everybody', '9', '9.0'],
  ]


def test_export_to_parquet(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  dataset.compute_signal(TestSignal(), 'text')

  # Download all columns.
  filepath = tmp_path / 'dataset.parquet'
  dataset.to_parquet(filepath)

  df = pd.read_parquet(filepath)
  expected_df = pd.DataFrame(
    [
      {
        'text': 'hello',
        'text.test_signal.len': np.int32(5),
        'text.test_signal.flen': np.float32(5.0),
      },
      {
        'text': 'everybody',
        'text.test_signal.len': np.int32(9),
        'text.test_signal.flen': np.float32(9.0),
      },
    ]
  )
  pd.testing.assert_frame_equal(df, expected_df)


def test_export_to_pandas(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data([{'text': 'hello'}, {'text': 'everybody'}])
  dataset.compute_signal(TestSignal(), 'text')

  # Download all columns.
  df = dataset.to_pandas()
  expected_df = pd.DataFrame(
    [
      {
        'text': 'hello',
        'text.test_signal.len': np.int32(5),
        'text.test_signal.flen': np.float32(5.0),
      },
      {
        'text': 'everybody',
        'text.test_signal.len': np.int32(9),
        'text.test_signal.flen': np.float32(9.0),
      },
    ]
  )
  pd.testing.assert_frame_equal(df, expected_df)

  # Select only some columns, including pseudocolumn rowid.
  df = dataset.to_pandas([ROWID, 'text', 'text.test_signal.flen'])
  expected_df = pd.DataFrame(
    [
      {ROWID: '1', 'text': 'hello', 'text.test_signal.flen': np.float32(5.0)},
      {ROWID: '2', 'text': 'everybody', 'text.test_signal.flen': np.float32(9.0)},
    ]
  )
  pd.testing.assert_frame_equal(df, expected_df)

  # Invalid columns.
  with pytest.raises(ValueError, match="Unable to select path \\('text', 'test_signal2'\\)"):
    dataset.to_pandas(['text', 'text.test_signal2'])


TEST_TIME = datetime(2023, 8, 15, 1, 23, 45)


@freeze_time(TEST_TIME)
def test_label_and_export_by_excluding(
  make_test_data: TestDataMaker, tmp_path: pathlib.Path
) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}, {'text': 'c'}])
  dataset.delete_rows(['2', '3'])

  # Download all, except the 'deleted' label.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath, exclude_labels=[DELETED_LABEL_NAME])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [
    {f'{DELETED_LABEL_NAME}.created': None, f'{DELETED_LABEL_NAME}.label': None, 'text': 'a'}
  ]

  # Download only the 'deleted' label.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath, include_labels=[DELETED_LABEL_NAME])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [
    {
      f'{DELETED_LABEL_NAME}.created': str(TEST_TIME),
      f'{DELETED_LABEL_NAME}.label': 'true',
      'text': 'b',
    },
    {
      f'{DELETED_LABEL_NAME}.created': str(TEST_TIME),
      f'{DELETED_LABEL_NAME}.label': 'true',
      'text': 'c',
    },
  ]


def test_include_multiple_labels(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}, {'text': 'c'}, {'text': 'd'}])
  dataset.add_labels('good', ['2', '3'])
  dataset.add_labels('very_good', ['3', '4'])

  # Include good and very_good when we export.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath, columns=['text'], include_labels=['good', 'very_good'])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  parsed_items = sorted(parsed_items, key=lambda x: x['text'])
  assert parsed_items == [{'text': 'b'}, {'text': 'c'}, {'text': 'd'}]


def test_exclude_multiple_labels(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}, {'text': 'c'}, {'text': 'd'}])
  dataset.add_labels('bad', ['2'])
  dataset.add_labels('very_bad', ['2', '3'])

  # Include good and very_good when we export.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath, columns=['text'], exclude_labels=['bad', 'very_bad'])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  parsed_items = sorted(parsed_items, key=lambda x: x['text'])
  assert parsed_items == [{'text': 'a'}, {'text': 'd'}]


def test_exclude_trumps_include(make_test_data: TestDataMaker, tmp_path: pathlib.Path) -> None:
  dataset = make_test_data([{'text': 'a'}, {'text': 'b'}, {'text': 'c'}, {'text': 'd'}])
  dataset.add_labels('good', ['2', '3', '4'])
  dataset.add_labels('bad', ['3', '4'])

  # Include good and very_good when we export.
  filepath = tmp_path / 'dataset.json'
  dataset.to_json(filepath, columns=['text'], include_labels=['good'], exclude_labels=['bad'])

  with open(filepath) as f:
    parsed_items = [json.loads(line) for line in f.readlines()]

  assert parsed_items == [{'text': 'b'}]



"""Tests for dataset.select_rows(filters=[...])."""

from typing import ClassVar, Iterable, Iterator, Optional

from ..schema import ROWID, Field, Item, MapType, RichData, field, schema
from ..signal import TextSignal
from .dataset import BinaryFilterTuple, ListFilterTuple, StringFilterTuple, UnaryFilterTuple
from .dataset_test_utils import TestDataMaker

TEST_DATA: list[Item] = [
  {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
  {'float': float('nan')},
]


STRING_TEST_DATA: list[Item] = [
  {'str': 'abcde', 'int': 1},
  {'str': 'a' * 10, 'int': 2},
  {'str': '', 'int': 3},
]


def test_filter_by_ids(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  id_filter: BinaryFilterTuple = (ROWID, 'equals', '1')
  result = dataset.select_rows(filters=[id_filter])

  assert list(result) == [{'str': 'a', 'int': 1, 'bool': False, 'float': 3.0}]

  id_filter = (ROWID, 'equals', '2')
  result = dataset.select_rows(filters=[id_filter])

  assert list(result) == [{'str': 'b', 'int': 2, 'bool': True, 'float': 2.0}]

  id_filter = (ROWID, 'equals', b'f')
  result = dataset.select_rows(filters=[id_filter])

  assert list(result) == []


def test_filter_deleted_interaction(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  id_filter: BinaryFilterTuple = (ROWID, 'equals', '1')
  result = dataset.select_rows(filters=[id_filter])

  assert list(result) == [{'str': 'a', 'int': 1, 'bool': False, 'float': 3.0}]

  dataset.delete_rows(filters=[id_filter])
  result = dataset.select_rows(filters=[id_filter])

  assert list(result) == []


def test_filter_greater(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: BinaryFilterTuple = ('float', 'greater', 2.0)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': 'a', 'int': 1, 'bool': False, 'float': 3.0}]


def test_filter_greater_equal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: BinaryFilterTuple = ('float', 'greater_equal', 2.0)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [
    {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
    {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  ]


def test_filter_less(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: BinaryFilterTuple = ('float', 'less', 2.0)
  result = dataset.select_rows(['*'], filters=[filter])

  assert list(result) == [{'str': 'b', 'int': 2, 'bool': True, 'float': 1.0}]


def test_filter_less_equal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: BinaryFilterTuple = ('float', 'less_equal', 2.0)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [
    {'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
    {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
  ]


def test_filter_not_equal(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: BinaryFilterTuple = ('float', 'not_equal', 2.0)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [
    {'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
    {'str': 'b', 'int': 2, 'bool': True, 'float': 1.0},
    # NaNs are not counted when we are filtering a field.
  ]


def test_filter_length_longer(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(STRING_TEST_DATA)

  filter: StringFilterTuple = ('str', 'length_longer', 5)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': 'abcde', 'int': 1}, {'str': 'a' * 10, 'int': 2}]


def test_filter_length_shorter(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(STRING_TEST_DATA)

  filter: StringFilterTuple = ('str', 'length_shorter', 5)
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': 'abcde', 'int': 1}, {'str': '', 'int': 3}]


def test_filter_string_regex_matches(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(STRING_TEST_DATA)

  filter: StringFilterTuple = ('str', 'regex_matches', 'a+')
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': 'abcde', 'int': 1}, {'str': 'a' * 10, 'int': 2}]


def test_filter_string_regex_matches_escaped(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'str': '..', 'int': 1},
      {'str': 'abc', 'int': 2},
    ]
  )

  filter: StringFilterTuple = ('str', 'regex_matches', r'\.+')
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': '..', 'int': 1}]


def test_filter_string_regex_matches_newlines(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(
    [
      {'str': 'hello\nworld', 'int': 1},
      {'str': 'hello\nworld\n', 'int': 2},
    ]
  )

  filter: StringFilterTuple = ('str', 'regex_matches', r'world\n')
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': 'hello\nworld\n', 'int': 2}]


def test_filter_string_not_regex_matches(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(STRING_TEST_DATA)

  filter: StringFilterTuple = ('str', 'not_regex_matches', 'a+')
  result = dataset.select_rows(filters=[filter])

  assert list(result) == [{'str': '', 'int': 3}]


def test_filter_by_list_of_ids(make_test_data: TestDataMaker) -> None:
  dataset = make_test_data(TEST_DATA)

  filter: ListFilterTuple = (ROWID, 'in', ['1', '2'])
  result = dataset.select_rows(['*', ROWID], filters=[filter])

  assert list(result) == [
    {ROWID: '1', 'str': 'a', 'int': 1, 'bool': False, 'float': 3.0},
    {ROWID: '2', 'str': 'b', 'int': 2, 'bool': True, 'float': 2.0},
  ]


def test_filter_by_exists(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'A', 'info': {'lang': 'en'}, 'ages': []},
    {'info': {'lang': 'fr'}},
    {'name': 'C', 'ages': [[1, 2], [3, 4]]},
  ]
  dataset = make_test_data(items)

  exists_filter: UnaryFilterTuple = ('name', 'exists')
  result = dataset.select_rows(['name'], filters=[exists_filter])
  assert list(result) == [{'name': 'A'}, {'name': 'C'}]

  exists_filter = ('info.lang', 'exists')
  result = dataset.select_rows(['name'], filters=[exists_filter])
  assert list(result) == [{'name': 'A'}, {'name': None}]

  exists_filter = ('ages.*.*', 'exists')
  result = dataset.select_rows(['name'], filters=[exists_filter])
  assert list(result) == [{'name': 'C'}]

  result = dataset.select_rows(['name'], filters=[('info', 'exists')])
  assert list(result) == [{'name': 'A'}, {'name': None}]


def test_filter_by_exists_on_enriched(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'A', 'info': {'lang': 'en'}, 'ages': []},
    {'info': {'lang': 'fr'}},
    {'name': 'C', 'ages': [[1, 2], [3, 4]]},
  ]
  dataset = make_test_data(items)

  class LengthSignal(TextSignal):
    name: ClassVar[str] = 'length_signal'

    def fields(self) -> Field:
      return field('int32')

    def compute(self, data: Iterable[RichData]) -> Iterator[Optional[Item]]:
      for text_content in data:
        yield len(text_content)

  dataset.compute_signal(LengthSignal(), path=('info', 'lang'))
  result = dataset.select_rows(['name'], filters=[('info', 'exists')])
  assert list(result) == [{'name': 'A'}, {'name': None}]


def test_filter_by_not_exists(make_test_data: TestDataMaker) -> None:
  items: list[Item] = [
    {'name': 'A', 'info': {'lang': 'en'}, 'ages': []},
    {'info': {'lang': 'fr'}},
    {'name': 'C', 'ages': [[1, 2], [3, 4]]},
  ]
  dataset = make_test_data(items)

  not_exists_filter: UnaryFilterTuple = ('name', 'not_exists')
  result = dataset.select_rows(['info'], filters=[not_exists_filter])
  assert list(result) == [{'info': {'lang': 'fr'}}]

  not_exists_filter = ('info.lang', 'not_exists')
  result = dataset.select_rows(['name'], filters=[not_exists_filter])
  assert list(result) == [{'name': 'C'}]

  not_exists_filter = ('ages', 'not_exists')
  result = dataset.select_rows(['name'], filters=[not_exists_filter])
  assert list(result) == [{'name': 'A'}, {'name': None}]

  result = dataset.select_rows(['name'], filters=[('info', 'not_exists')])
  assert list(result) == [{'name': 'C'}]


def test_map_dtype(make_test_data: TestDataMaker) -> None:
  items = [
    {'column': {'a': 1.0, 'b': 2.0}},
    {'column': {'b': 2.5}},
    {'column': {'a': 3.0, 'c': 3.5}},
  ]
  map_dtype = MapType(key_type='string', value_field=field('float32'))
  data_schema = schema({'column': Field(dtype=map_dtype)})
  dataset = make_test_data(items, schema=data_schema)

  result = dataset.select_rows(['column'], filters=[('column.a', 'greater', 2.0)])
  assert list(result) == [
    {'column': {'key': ['a', 'c'], 'value': [3.0, 3.5]}},
  ]

  result = dataset.select_rows(['column'], filters=[('column.b', 'less_equal', 2.0)])
  assert list(result) == [{'column': {'key': ['a', 'b'], 'value': [1.0, 2.0]}}]

  result = dataset.select_rows(['column'], filters=[('column.a', 'exists')])
  assert list(result) == [
    {'column': {'key': ['a', 'b'], 'value': [1.0, 2.0]}},
    {'column': {'key': ['a', 'c'], 'value': [3.0, 3.5]}},
  ]



"""Find secret keys in text.

# Code forked from
# https://github.com/bigcode-project/pii-lib/blob/main/utils/keys_detection.py
# under the Apache 2.0 License.
"""
import os
import tempfile
from typing import Iterator

from detect_secrets import SecretsCollection
from detect_secrets.settings import transient_settings

from ..schema import Item, span

# Secrets detection with detect-secrets tool

filters = [
  # some filters from
  # https://github.com/Yelp/detect-secrets/blob/master/docs/filters.md#built-in-filters
  # were removed based on their targets
  {'path': 'detect_secrets.filters.heuristic.is_potential_uuid'},
  {'path': 'detect_secrets.filters.heuristic.is_likely_id_string'},
  {'path': 'detect_secrets.filters.heuristic.is_templated_secret'},
  {'path': 'detect_secrets.filters.heuristic.is_sequential_string'},
]
plugins = [
  {'name': 'ArtifactoryDetector'},
  {'name': 'AWSKeyDetector'},
  {'name': 'AzureStorageKeyDetector'},
  {'name': 'CloudantDetector'},
  {'name': 'DiscordBotTokenDetector'},
  {'name': 'GitHubTokenDetector'},
  {'name': 'IbmCloudIamDetector'},
  {'name': 'IbmCosHmacDetector'},
  {'name': 'JwtTokenDetector'},
  {'name': 'MailchimpDetector'},
  {'name': 'NpmDetector'},
  {'name': 'SendGridDetector'},
  {'name': 'SlackDetector'},
  {'name': 'SoftlayerDetector'},
  {'name': 'StripeDetector'},
  {'name': 'TwilioKeyDetector'},
  # remove 3 plugins for keyword
  # {'name': 'BasicAuthDetector'},
  # {'name': 'KeywordDetector'},
  # {'name': 'PrivateKeyDetector'},
]


def _is_hash(content: str, value: str) -> bool:
  """Second check if the value is a hash (after gibberish detector)."""
  try:
    index = content.index(value)
  except ValueError:
    return False
  lines = content[:index].splitlines()
  if not lines:
    return False
  target_line = lines[-1]
  if len(value) in [32, 40, 64]:
    # if 'sha' or 'md5' are in content:
    keywords = ['sha', 'md5', 'hash', 'byte']
    if any(x in target_line.lower() for x in keywords):
      return True
  return False


def _file_has_hashes(content: str, coeff: float = 0.02) -> bool:
  """Checks if the file contains literals 'hash' or 'sha' for more than 2% nb_of_lines."""
  lines = content.splitlines()
  count_sha = 0
  count_hash = 0
  nlines = content.count('\n')
  threshold = int(coeff * nlines)
  for line in lines:
    count_sha += line.lower().count('sha')
    count_hash += line.lower().count('hash')
    if count_sha > threshold or count_hash > threshold:
      return True
  return False


def _get_indexes(text: str, value: str) -> list[tuple[int, int]]:
  string = text
  indexes: list[int] = []
  new_start = 0
  while True:
    try:
      start = string.index(value)
      indexes.append(new_start + start)
      new_start = new_start + start + len(value)
      string = text[new_start:]
    except ValueError:
      break
  return [(x, x + len(value)) for x in indexes]


def find_secrets(content: str, suffix: str = '.txt') -> Iterator[Item]:
  """Detect secret keys in content using detect-secrets tool."""
  fp = tempfile.NamedTemporaryFile(suffix=suffix, delete=False, mode='w', encoding='utf-8')
  fp.write(content)
  fp.close()
  secrets = SecretsCollection()
  with transient_settings({'plugins_used': plugins, 'filters_used': filters}):
    secrets.scan_file(fp.name)
  os.unlink(fp.name)
  secrets_set = list(secrets.data.values())
  if not secrets_set:
    return
  for secret in secrets_set[0]:
    if not secret.secret_value:
      continue
    if _is_hash(content, secret.secret_value) or _file_has_hashes(content):
      continue
    indexes = _get_indexes(content, secret.secret_value)
    for start, end in indexes:
      yield span(start, end)



"""A data loader standalone binary. This should only be run as a script to load a dataset.

To run the source loader as a binary directly:

poetry run python -m lilac.load_dataset \
  --dataset_name=movies_dataset \
  --output_dir=./data/ \
  --config_path=./datasets/the_movies_dataset.json
"""
import os
import pathlib
import uuid
from typing import Iterable, Iterator, Optional, Union, cast

import pandas as pd
from datasets import Dataset as HFDataset
from datasets import DatasetDict as HFDatasetDict
from datasets import load_dataset

from .config import DatasetConfig
from .data.dataset import Dataset, default_settings
from .data.dataset_utils import write_items_to_parquet
from .db_manager import get_dataset
from .env import get_project_dir
from .project import add_project_dataset_config, update_project_dataset_settings
from .schema import MANIFEST_FILENAME, PARQUET_FILENAME_PREFIX, ROWID, Field, Item, Schema, is_float
from .source import Source, SourceManifest
from .sources.dict_source import DictSource
from .sources.huggingface_source import HuggingFaceSource
from .tasks import TaskId, get_progress_bar
from .utils import get_dataset_output_dir, log, open_file


def create_dataset(
  config: DatasetConfig,
  project_dir: Optional[Union[str, pathlib.Path]] = None,
  overwrite: bool = False,
) -> Dataset:
  """Load a dataset from a given source configuration.

  Args:
    config: The dataset configuration to load.
    project_dir: The path to the project directory for where to create the dataset. If not defined,
      uses the project directory from `LILAC_PROJECT_DIR` or [deprecated] `LILAC_DATA_PATH`.
    overwrite: Whether to overwrite the dataset if it already exists.
  """
  project_dir = project_dir or get_project_dir()
  if not project_dir:
    raise ValueError(
      '`project_dir` must be defined. Please pass a `project_dir` or set it '
      'globally with `set_project_dir(path)`'
    )
  add_project_dataset_config(config, project_dir, overwrite)

  process_source(project_dir, config)

  return get_dataset(config.namespace, config.name, project_dir)


def from_dicts(
  namespace: str, name: str, items: Iterable[Item], overwrite: bool = False
) -> Dataset:
  """Load a dataset from an iterable of python dictionaries."""
  config = DatasetConfig(
    namespace=namespace,
    name=name,
    source=DictSource(items),
  )
  return create_dataset(config, overwrite=overwrite)


def from_huggingface(
  dataset: Union[str, HFDataset, HFDatasetDict],
  namespace: str = 'local',
  name: Optional[str] = None,
  overwrite: bool = False,
) -> Dataset:
  """Load a dataset from HuggingFace.

  Args:
    dataset: A HuggingFace dataset or its name registered on the hub.
    namespace: The Lilac namespace for the loaded dataset. Defaults to `local`.
    name: The Lilac name of the dataset to create. Defaults to the name of the HuggingFace dataset.
    overwrite: Whether to overwrite the dataset if it already exists.
  """
  if isinstance(dataset, str):
    dataset = cast(Union[HFDataset, HFDatasetDict], load_dataset(dataset))

  if not namespace:
    namespace = 'local'

  if not name:
    # Infer name from the dataset instance.
    split: HFDataset = dataset
    if isinstance(dataset, HFDatasetDict):
      split = list(cast(HFDatasetDict, dataset).values())[0]
    if split.info.dataset_name:
      name = split.info.dataset_name

  if not name:
    raise ValueError('`name` must be defined since it could not be inferred from the dataset.')

  config = DatasetConfig(
    namespace=namespace,
    name=name,
    source=HuggingFaceSource(dataset_name=name, dataset=dataset),
  )
  return create_dataset(config, overwrite=overwrite)


def process_source(
  project_dir: Union[str, pathlib.Path],
  config: DatasetConfig,
  task_id: Optional[TaskId] = None,
) -> str:
  """Process a source."""
  output_dir = get_dataset_output_dir(project_dir, config.namespace, config.name)

  config.source.setup()
  try:
    manifest = config.source.load_to_parquet(output_dir, task_id=task_id)
  except NotImplementedError:
    manifest = slow_process(config.source, output_dir, task_id=task_id)

  with open_file(os.path.join(output_dir, MANIFEST_FILENAME), 'w') as f:
    f.write(manifest.model_dump_json(indent=2, exclude_none=True))

  if not config.settings:
    dataset = get_dataset(config.namespace, config.name, project_dir)
    settings = default_settings(dataset)
    update_project_dataset_settings(config.namespace, config.name, settings, project_dir)

  log(f'Dataset "{config.name}" written to {output_dir}')

  return output_dir


def slow_process(
  source: Source, output_dir: str, task_id: Optional[TaskId] = None
) -> SourceManifest:
  """Process the items of a source, writing to parquet.

  Args:
    source: The source to process.
    output_dir: The directory to write the parquet files to.
    task_id: The TaskManager id for this process run. This is used to update the task progress.

  Returns:
    A SourceManifest that describes schema and parquet file locations.
  """
  source_schema = source.source_schema()
  items = source.yield_items()
  # Add rowids and fix NaN in string columns.
  items = normalize_items(items, source_schema.fields)

  # Add progress.
  progress_bar = get_progress_bar(
    task_id=task_id,
    estimated_len=source_schema.num_items,
  )
  items = progress_bar(items)

  # Filter out the `None`s after progress.
  items = (item for item in items if item is not None)

  data_schema = Schema(fields=source_schema.fields.copy())
  filepath = write_items_to_parquet(
    items=items,
    output_dir=output_dir,
    schema=data_schema,
    filename_prefix=PARQUET_FILENAME_PREFIX,
    shard_index=0,
    num_shards=1,
  )

  filenames = [os.path.basename(filepath)]
  manifest = SourceManifest(files=filenames, data_schema=data_schema, images=None, source=source)
  return manifest


def normalize_items(items: Iterable[Item], fields: dict[str, Field]) -> Iterator[Item]:
  """Sanitize items by removing NaNs and NaTs."""
  replace_nan_fields = [
    field_name for field_name, field in fields.items() if field.dtype and not is_float(field.dtype)
  ]
  for item in items:
    if item is None:
      yield item
      continue

    # Add rowid if it doesn't exist.
    if ROWID not in item:
      item[ROWID] = uuid.uuid4().hex

    # Fix NaN values.
    for field_name in replace_nan_fields:
      item_value = item.get(field_name)
      if item_value and not isinstance(item_value, Iterable) and pd.isna(item_value):
        item[field_name] = None

    yield item



"""A source that reads from an Iterable of LlamaIndex Documents."""
from typing import TYPE_CHECKING, Any, ClassVar, Iterable, Iterator, Optional

if TYPE_CHECKING:
  from llama_index import Document

from typing_extensions import override

from ..schema import Item
from ..source import Source, SourceSchema
from .dict_source import DictSource


class LlamaIndexDocsSource(Source):
  """LlamaIndex document source

  Loads documents from a LlamaIndex Document Iterable.

  Usage:
  ```python
    from llama_index import download_loader

    # See: https://llamahub.ai/l/papers-arxiv
    ArxivReader = download_loader("ArxivReader")

    loader = ArxivReader()
    documents = loader.load_data(search_query='au:Karpathy')

    # Create the dataset
    ll.create_dataset(
    config=ll.DatasetConfig(
      namespace='local',
      name='arxiv-karpathy',
      source=ll.LlamaIndexDocsSource(
        # documents comes from the loader.load_data call in the previous cell.
        documents=documents
        )
      )
    )
  ```

  A detailed example notebook
  [can be found here](https://github.com/lilacai/lilac/blob/main/notebooks/LlamaIndexLoader.ipynb)
  """  # noqa: D415, D400

  name: ClassVar[str] = 'llama_index_docs'

  _documents: Optional[Iterable['Document']]
  # Used to infer the schema.
  _infer_schema_docs: Iterator['Document']

  _dict_source: Optional[DictSource] = None

  def __init__(self, documents: Optional[Iterable['Document']] = None, **kwargs: Any):
    super().__init__(**kwargs)
    self._documents = documents

  @override
  def setup(self) -> None:
    """Setup the source."""
    if not self._documents:
      raise ValueError('cls argument `documents` is not defined.')

    self._dict_source = DictSource(self._get_doc_items())
    self._dict_source.setup()

  def _get_doc_items(self) -> Iterable[Item]:
    if not self._documents:
      raise ValueError('cls argument `documents` is not defined.')

    for doc in self._documents:
      yield {'doc_id': str(doc.doc_id), 'text': doc.text, **doc.metadata}

  @override
  def source_schema(self) -> SourceSchema:
    """Return the source schema."""
    if not self._dict_source:
      raise ValueError('Please call setup() before calling `source_schema`.')

    return self._dict_source.source_schema()

  @override
  def yield_items(self) -> Iterable[Item]:
    """Ingest the documents."""
    if not self._dict_source:
      raise ValueError('Please call setup() before calling `process`.')

    return self._dict_source.yield_items()



"""Huggingface source."""
import multiprocessing
import os
from typing import Any, ClassVar, Optional, Union, cast

import duckdb
from datasets import (
  ClassLabel,
  Dataset,
  DatasetDict,
  Image,
  Sequence,
  Translation,
  Value,
  load_dataset,
  load_from_disk,
)
from pydantic import BaseModel, ConfigDict, GetJsonSchemaHandler
from pydantic import Field as PydanticField
from pydantic_core import CoreSchema
from typing_extensions import override

from ..data import dataset_utils
from ..schema import (
  INT32,
  PARQUET_FILENAME_PREFIX,
  ROWID,
  STRING,
  Field,
  Schema,
  arrow_dtype_to_dtype,
)
from ..source import Source, SourceManifest, SourceSchema
from ..tasks import TaskId
from ..utils import log

HF_SPLIT_COLUMN = '__hfsplit__'

# Used when the dataset is saved locally.
DEFAULT_LOCAL_SPLIT_NAME = 'default'


class SchemaInfo(BaseModel):
  """Information about the processed huggingface schema."""

  fields: dict[str, Field] = {}
  class_labels: dict[str, list[str]]
  num_items: int


def _infer_field(feature_value: Union[Value, dict]) -> Optional[Field]:
  """Infer the field type from the feature value."""
  if isinstance(feature_value, dict):
    fields: dict[str, Field] = {}
    for name, value in feature_value.items():
      field = _infer_field(value)
      if field:
        fields[name] = field
    return Field(fields=fields)
  elif isinstance(feature_value, Value):
    return Field(dtype=arrow_dtype_to_dtype(feature_value.pa_type))
  elif isinstance(feature_value, Sequence):
    # Huggingface Sequences can contain a dictionary of feature values, e.g.
    #   Sequence(feature={'x': Value(dtype='int32'), 'y': Value(dtype='float32')}}
    # These are converted to {'x': [...]} and {'y': [...]}
    if isinstance(feature_value.feature, dict):
      return Field(
        fields={
          name: Field(repeated_field=_infer_field(value))
          for name, value in feature_value.feature.items()
        }
      )
    else:
      return Field(repeated_field=_infer_field(feature_value.feature))
  elif isinstance(feature_value, list):
    if len(feature_value) > 1:
      raise ValueError('Field arrays with multiple values are not supported.')
    return Field(repeated_field=_infer_field(feature_value[0]))
  elif isinstance(feature_value, ClassLabel):
    # TODO(nsthorat): For nested class labels, return the path with the class label values to show
    # strings in the UI.
    return Field(dtype=INT32)
  elif isinstance(feature_value, Image):
    log(f'{feature_value} has type Image and is ignored.')
    return None
  else:
    raise ValueError(f'Feature is not a `Value`, `Sequence`, or `dict`: {feature_value}')


def hf_schema_to_schema(
  hf_dataset_dict: DatasetDict, split: Optional[str], sample_size: Optional[int]
) -> SchemaInfo:
  """Convert the HuggingFace schema to our schema."""
  if split:
    split_datasets = [hf_dataset_dict[split]]
  else:
    split_datasets = [hf_dataset_dict[split] for split in hf_dataset_dict.keys()]

  fields: dict[str, Field] = {}
  class_labels: dict[str, list[str]] = {}
  num_items = 0

  for split_dataset in split_datasets:
    split_size = len(split_dataset)
    if sample_size:
      split_size = min(split_size, sample_size)
    num_items += split_size

    features = split_dataset.features
    for feature_name, feature_value in features.items():
      if feature_name in fields:
        continue

      if isinstance(feature_value, ClassLabel):
        # Class labels act as strings and we map the integer to a string before writing.
        fields[feature_name] = Field(dtype=STRING)
        class_labels[feature_name] = feature_value.names
      elif isinstance(feature_value, Translation):
        # Translations act as categorical strings.
        language_fields: dict[str, Field] = {}
        for language in feature_value.languages:
          language_fields[language] = Field(dtype=STRING)
        fields[feature_name] = Field(fields=language_fields)
      else:
        field = _infer_field(feature_value)
        if field:
          fields[feature_name] = field

  # Add the split column to the schema.
  fields[HF_SPLIT_COLUMN] = Field(dtype=STRING)

  return SchemaInfo(fields=fields, class_labels=class_labels, num_items=num_items)


class HuggingFaceSource(Source):
  """HuggingFace data loader

  For a list of datasets see: [huggingface.co/datasets](https://huggingface.co/datasets).

  For documentation on dataset loading see:
      [huggingface.co/docs/datasets/index](https://huggingface.co/docs/datasets/index)
  """  # noqa: D415, D400

  name: ClassVar[str] = 'huggingface'

  dataset_name: str = PydanticField(description='Either in the format `user/dataset` or `dataset`.')
  dataset: Optional[Union[Dataset, DatasetDict]] = PydanticField(
    description='The pre-loaded HuggingFace dataset', exclude=True, default=None
  )
  config_name: Optional[str] = PydanticField(
    title='Dataset config name', description='Some datasets require this.', default=None
  )
  split: Optional[str] = PydanticField(
    title='Dataset split', description='Loads all splits by default.', default=None
  )
  sample_size: Optional[int] = PydanticField(
    title='Sample size',
    description='Number of rows to sample from the dataset, for each split.',
    default=None,
  )
  token: Optional[str] = PydanticField(
    title='Huggingface token',
    description='Huggingface token for private datasets.',
    default=None,
    exclude=True,
  )
  revision: Optional[str] = PydanticField(title='Dataset revision', default=None)
  load_from_disk: Optional[bool] = PydanticField(
    description='Load from local disk instead of the hub.', default=False
  )
  _dataset_dict: Optional[DatasetDict] = None
  _schema_info: Optional[SchemaInfo] = None
  model_config = ConfigDict(arbitrary_types_allowed=True)

  # We override this method to remove the `dataset_dict` from the json schema.
  @classmethod
  def __get_pydantic_json_schema__(
    cls, core_schema: CoreSchema, handler: GetJsonSchemaHandler
  ) -> dict[str, Any]:
    fields = cast(dict, core_schema)['schema']['fields']
    if 'dataset' in fields:
      del fields['dataset']
    json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
    json_schema = handler.resolve_ref_schema(json_schema)
    return json_schema

  @override
  def setup(self) -> None:
    if self.dataset:
      if isinstance(self.dataset, Dataset):
        self._dataset_dict = DatasetDict({DEFAULT_LOCAL_SPLIT_NAME: self.dataset})
      elif isinstance(self.dataset, DatasetDict):
        self._dataset_dict = self.dataset
      else:
        raise ValueError(f'Unsupported dataset type: {type(self.dataset)}')
    else:
      if self.load_from_disk:
        # Load from disk.
        hf_dataset_dict = load_from_disk(self.dataset_name)
        if isinstance(hf_dataset_dict, Dataset):
          hf_dataset_dict = DatasetDict({DEFAULT_LOCAL_SPLIT_NAME: hf_dataset_dict})
      else:
        hf_dataset_dict = load_dataset(
          self.dataset_name,
          self.config_name,
          num_proc=multiprocessing.cpu_count(),
          verification_mode='no_checks',
          token=self.token,
        )
      self._dataset_dict = hf_dataset_dict
    self._schema_info = hf_schema_to_schema(self._dataset_dict, self.split, self.sample_size)

  @override
  def source_schema(self) -> SourceSchema:
    if not self._schema_info:
      raise ValueError('`setup()` must be called before `source_schema`.')
    return SourceSchema(fields=self._schema_info.fields, num_items=self._schema_info.num_items)

  @override
  def load_to_parquet(self, output_dir: str, task_id: Optional[TaskId]) -> SourceManifest:
    del task_id
    if not self._schema_info or not self._dataset_dict:
      raise ValueError('`setup()` must be called before `process`.')

    out_filename = dataset_utils.get_parquet_filename(PARQUET_FILENAME_PREFIX, 0, 1)
    filepath = os.path.join(output_dir, out_filename)
    os.makedirs(output_dir, exist_ok=True)

    if self.split:
      split_names = [self.split]
    else:
      split_names = list(self._dataset_dict.keys())

    # DuckDB uses the locals() namespace to enable chaining SQL queries based on Python variables as
    # table names. Since we have an unknown number of splits, we dynamically create names in
    # locals() for DuckDB to refererence.
    duckdb_splits_local_vars = {}
    for split_name in split_names:
      duckdb_handle = f'_duckdb_handle_{split_name}'
      duckdb_splits_local_vars[split_name] = duckdb_handle
      locals()[duckdb_handle] = self._dataset_dict[split_name].data.table
    sql_query = '\nUNION ALL\n'.join(
      f"""(SELECT *, '{split_name}' AS {HF_SPLIT_COLUMN}
        FROM {duckdb_handle}
        {f'LIMIT {self.sample_size}' if self.sample_size else ''})"""
      for split_name, duckdb_handle in duckdb_splits_local_vars.items()
    )
    all_splits = duckdb.sql(sql_query)

    # Enum column rewrites. (Huggingface supports integer columns, with a schema mapping integers
    # to strings. We rewrite the integer columns to strings.)
    if self._schema_info.class_labels:
      star_clause = f'* EXCLUDE ({", ".join(self._schema_info.class_labels.keys())}),'
      # +1 because HF class labels start at 0 but DuckDB's array_extract is 1-indexed.
      rename_clause = ', '.join(
        f'ARRAY_EXTRACT({class_names}, {key} + 1) AS {key}'
        for key, class_names in self._schema_info.class_labels.items()
      )
      select_clause = star_clause + rename_clause
    else:
      select_clause = '*'
    duckdb.sql(
      f"""
      SELECT replace(CAST(uuid() AS VARCHAR), '-', '') AS {ROWID},
      {select_clause}
      FROM all_splits
      """
    ).write_parquet(filepath, compression='zstd')
    schema = Schema(fields=self.source_schema().fields.copy())
    return SourceManifest(files=[out_filename], data_schema=schema, source=self)



"""Registers all available dataset formats."""

from ..dataset_format import register_dataset_format
from .openchat import OpenChat
from .sharegpt import ShareGPT


def register_default_formats() -> None:
  """Register all the default dataset formats."""
  register_dataset_format(ShareGPT)
  register_dataset_format(OpenChat)



"""Tests for the language detection signal."""

from pytest_mock import MockerFixture

from ..schema import span
from . import lang_detection
from .lang_detection import LANG_CODE, LangDetectionSignal


def test_lang_detection_sentences(mocker: MockerFixture) -> None:
  signal = LangDetectionSignal()
  mocker.patch(f'{lang_detection.__name__}.TEXT_LEN_THRESHOLD', 1)
  signal.setup()
  docs = ['War doesnt show whos right, just whos left.', 'Ein, zwei, drei, vier']
  res = list(signal.compute(docs))
  assert res == ['en', 'de']


def test_lang_detection_multiple_paragraphs(mocker: MockerFixture) -> None:
  signal = LangDetectionSignal(split_by_paragraph=True)
  mocker.patch(f'{lang_detection.__name__}.TEXT_LEN_THRESHOLD', 1)
  signal.setup()
  doc = 'War doesnt show whos right, just whos left.\n\nEin, zwei, drei, vier'
  res = list(signal.compute([doc]))
  assert res == [[span(0, 43, {LANG_CODE: 'en'}), span(45, 66, {LANG_CODE: 'de'})]]


def test_text_too_short(mocker: MockerFixture) -> None:
  signal = LangDetectionSignal()
  mocker.patch(f'{lang_detection.__name__}.TEXT_LEN_THRESHOLD', 25)
  signal.setup()
  docs = ['War doesnt show whos right, just whos left.', 'Ein, zwei, drei, vier']
  res = list(signal.compute(docs))
  assert res == ['en', 'TOO_SHORT']



"""Find near-duplicates using minhash.

# Code forked from
# https://github.com/bigcode-project/bigcode-dataset/blob/main/near_deduplication/minhash_deduplication.py
# under the Apache 2.0 License.
"""
import gc
import hashlib
import re
import struct
from collections import defaultdict
from itertools import tee
from typing import Iterable, List

import numpy as np
from tqdm import tqdm

SEED = 42
WHITESPACE = re.compile(r'\s+')
RNG = np.random.RandomState(SEED)
MAX_HASH = np.uint64((1 << 32) - 1)
MERSENNE_PRIME = np.uint64((1 << 61) - 1)


def _ngrams(sequence: List[str], n: int, min_ngram_size: int) -> Iterable:
  """Directly taken from nltk package to avoid dependency.

  Args:
    sequence: The sequence of items to be n-grammed.
    n: The order of the n-grams to be extracted.
    min_ngram_size: The minimum size of n-grams.

  Returns:
    The n-grams generated from the sequence.
  """
  if len(sequence) < min_ngram_size:
    return []
  ngram_size = min(n, len(sequence))
  iterables = tee(sequence, ngram_size)
  for i, sub_iterable in enumerate(iterables):
    for _ in range(i):
      next(sub_iterable, None)
  return zip(*iterables)


def _sha1_hash32(data: bytes) -> int:
  """Directly taken from datasketch package to avoid dependency."""
  return struct.unpack('<I', hashlib.sha1(data).digest()[:4])[0]


def _embed_func(
  content: str,
  num_perm: int,
  ngram_size: int,
  hashranges: list[tuple[int, int]],
  permutations: np.ndarray,
  min_ngram_size: int,
) -> list[bytes]:
  """Combined with some datasketch code to better parallelize computation.

  Args:
    content: The content to be embedded.
    idx: The index of the content.
    num_perm: The number of permutations.
    ngram_size: The size of n-grams.
    hashranges: The ranges of hash values.
    permutations: The permutations for the minhash.
    min_ngram_size: The minimum size of n-grams.

  Returns:
    The hash values in each range and the index.
  """
  hashvalues = np.ones(num_perm, dtype=np.uint64) * MAX_HASH
  tokens = {' '.join(t) for t in _ngrams(WHITESPACE.split(content), ngram_size, min_ngram_size)}
  hv = np.array([_sha1_hash32(token.encode('utf-8')) for token in tokens], dtype=np.uint64)  # noqa: E501
  a, b = permutations
  phv = np.bitwise_and(((hv * np.tile(a, (len(hv), 1)).T).T + b) % MERSENNE_PRIME, MAX_HASH)  # noqa: E501
  hashvalues = np.vstack([phv, hashvalues]).min(axis=0)
  Hs: list[bytes] = [bytes(hashvalues[start:end].byteswap().data) for start, end in hashranges]
  return Hs


def _optimal_param(
  threshold: float,
  num_perm: int,
  false_positive_weight: float = 0.5,
  false_negative_weight: float = 0.5,
) -> tuple[int, int]:
  """Find optimal `MinHashLSH` parameter that minimizes the weighted sum of false pos and false neg.

  Taken from datasketch.

  Args:
    threshold: The threshold for similarity.
    num_perm: The number of permutations.
    false_positive_weight: The weight of false positive.
    false_negative_weight: The weight of false negative.

  Returns:
    The optimal `b` and `r` parameters.
    The number of bands, and the number of rows per band respectively.
  """
  from scipy.integrate import quad as integrate

  def false_positive_probability(threshold: float, b: int, r: int) -> float:
    """Source: `datasketch.lsh`."""

    def proba(s: float) -> float:
      return 1 - (1 - s ** float(r)) ** float(b)

    a, _ = integrate(proba, 0.0, threshold)
    return a

  def false_negative_probability(threshold: float, b: int, r: int) -> float:
    """Source: `datasketch.lsh`."""

    def proba(s: float) -> float:
      return 1 - (1 - (1 - s ** float(r)) ** float(b))

    a, _ = integrate(proba, threshold, 1.0)
    return a

  min_error = float('inf')
  opt = (0, 0)
  for b in range(1, num_perm + 1):
    max_r = int(num_perm / b)
    for r in range(1, max_r + 1):
      fp = false_positive_probability(threshold, b, r)
      fn = false_negative_probability(threshold, b, r)
      error = fp * false_positive_weight + fn * false_negative_weight
      if error < min_error:
        min_error = error
        opt = (b, r)
  return opt


class UnionFind:
  """Union find data structure."""

  def __init__(self) -> None:
    self.parent: dict[int, int] = {}

  def find(self, x: int) -> int:
    """Find the parent of the node."""
    if x not in self.parent:
      self.parent[x] = x
    if self.parent[x] != x:
      self.parent[x] = self.find(self.parent[x])
    return self.parent[x]

  def union(self, x: int, y: int) -> None:
    """Union two nodes."""
    px = self.find(x)
    py = self.find(y)
    self.parent[px] = self.parent[py] = min(px, py)


def find_clusters(
  data: Iterable[str],
  ngram_size: int = 5,
  num_perm: int = 256,
  threshold: float = 0.7,
  min_ngram_size: int = 1,
) -> Iterable[int]:
  """Deduplicates documents and returns cluster ids."""
  uf = UnionFind()
  B, R = _optimal_param(threshold, num_perm)
  HASH_RANGES: list[tuple[int, int]] = [(i * R, (i + 1) * R) for i in range(B)]
  HASH_TABLES: list[dict[bytes, set[int]]] = [defaultdict(set) for _ in range(B)]

  # Consume the data.
  PERMUTATIONS = np.array(
    [
      (
        RNG.randint(1, MERSENNE_PRIME, dtype=np.uint64),
        RNG.randint(0, MERSENNE_PRIME, dtype=np.uint64),
      )
      for _ in range(num_perm)
    ],
    dtype=np.uint64,
  ).T

  # Fingerprinting.
  embedded: list[tuple[int, list[bytes]]] = []
  for key, content in tqdm(enumerate(data), dynamic_ncols=True, desc='Fingerprinting...'):
    hashes = _embed_func(
      content,
      num_perm=num_perm,
      hashranges=HASH_RANGES,
      ngram_size=ngram_size,
      permutations=PERMUTATIONS,
      min_ngram_size=min_ngram_size,
    )
    embedded.append((key, hashes))

  batch_size: int = 10000
  for i in tqdm(
    range(0, len(embedded), batch_size), dynamic_ncols=True, desc='Computing hash collisions...'
  ):
    batch = embedded[i : i + batch_size]
    for key, Hs in batch:
      for H, hashtable in zip(Hs, HASH_TABLES):
        hashtable[H].add(key)

  for table in tqdm(HASH_TABLES, dynamic_ncols=True, desc='Clustering...'):
    for cluster in table.values():
      if len(cluster) <= 1:
        continue
      idx = min(cluster)
      for x in cluster:
        uf.union(x, idx)

  gc.freeze()
  gc.disable()
  cluster_ids = [uf.find(i) for i in range(len(embedded))]
  gc.enable()
  gc.collect()

  return cluster_ids



"""Tests for the Llama Index Docs source."""

from llama_index import Document

from ..schema import schema
from ..source import SourceSchema
from .llama_index_docs_source import LlamaIndexDocsSource


def test_simple_llama_index_documents() -> None:
  docs = [
    Document(doc_id='id_1', text='test', extra_info={'name': 'a', 'age': 1}),
    Document(doc_id='id_2', text='test2', extra_info={'name': 'b', 'age': 2}),
    Document(doc_id='id_3', text='test3', extra_info={'name': 'c', 'age': 3}),
  ]

  source = LlamaIndexDocsSource(docs)
  source.setup()

  source_schema = source.source_schema()
  assert source_schema == SourceSchema(
    fields=schema({'doc_id': 'string', 'text': 'string', 'name': 'string', 'age': 'int64'}).fields,
    num_items=3,
  )

  items = list(source.yield_items())

  assert items == [
    {'doc_id': 'id_1', 'text': 'test', 'name': 'a', 'age': 1},
    {'doc_id': 'id_2', 'text': 'test2', 'name': 'b', 'age': 2},
    {'doc_id': 'id_3', 'text': 'test3', 'name': 'c', 'age': 3},
  ]



"""Tests the chunk splitter."""

from .chunk_splitter import split_text
from .text_splitter_test_utils import text_to_textchunk


def test_paragraphs_no_overlap() -> None:
  text = 'Hello.\n\nThis will get split.\n\nThe sentence\n\nA.\n\nB.\n\nC.'
  split_items = split_text(text, chunk_size=12, chunk_overlap=0)

  # "This will get split" should split in 2 chunks, and "A.\n\nB.\n\nC." should be 1 chunk.
  expected_spans = text_to_textchunk(
    text, ['Hello.', 'This will', 'get split.', 'The sentence', 'A.\n\nB.\n\nC.']
  )
  assert split_items == expected_spans


def test_single_world_is_too_long_no_overlap() -> None:
  text = 'ThisIsASingleWordThatIsTooLong'
  split_items = split_text(text, chunk_size=6, chunk_overlap=0)

  expected_spans = text_to_textchunk(text, ['ThisIs', 'ASingl', 'eWordT', 'hatIsT', 'ooLong'])
  assert split_items == expected_spans


def test_newlines_with_overlap() -> None:
  text = 'Hello.\n\nWorld.\n\nThis will get split.'
  split_items = split_text(text, chunk_size=12, chunk_overlap=5)

  expected_chunks = text_to_textchunk(
    text, ['Hello.', 'World.', 'This will', 'will get', 'get split.'], allowable_overlap=5
  )
  assert split_items == expected_chunks
